/* JPG: 1.9.91 */
/* determinate unfolding to first choice point, then first recursive goal
   with more specific clause transforms */

/****************************************************/
/* export:  charpath/3,                             */
/* export:  more_specific/1,                        */
/* export:  cleanup/3,                              */
/****************************************************/



resultant(A,R) :-
	resultant(A,_,R).


resultant(G,P,R2) :-
	repeat_msv_prefix(G,P1,[],R1,det),
	unfold_choicepoint(R1,R2,P2,P1),
	repeat_msv_prefix(R2,P,P2,_,det).
/*
	repeat_msv_prefix1(R2,P,[choice|P2],R3,det).
*/
	
repeat_msv_prefix(R,P,P,R,nondet) :-
	!.
repeat_msv_prefix(R,P1,P3,R1,det) :-
	more_specific(R),
	prefix(R,P2,P3,R2,Det),
	repeat_msv_prefix(R2,P1,P2,R1,Det).

repeat_msv_prefix1(R,P,P,R,nondet) :-
	!.
repeat_msv_prefix1(R,P1,P3,R1,det) :-
	more_specific(R),
	prefix1(R,P2,P3,R2,Det),
	repeat_msv_prefix1(R2,P1,P2,R1,Det).

more_specific(B) :-
	var(B),
	!.
more_specific((B,Bs)) :-
	!,
	more_specific(B),
	more_specific(Bs).
more_specific(B) :-
        solve_choice(B,_).

unfold_choicepoint((R,Rs),Bs,P,Q) :-
        !,
        unfold_choicepoint(R,R1,P,Q),
        joingoals(R1,Rs,Bs).
unfold_choicepoint(true,true,P,P) :-
        !.
unfold_choicepoint(R,R,P,P) :-
        imported(R),
        !.
unfold_choicepoint(R,true,[(1,N)|P],P) :-
        sp_builtin(R),
	unfoldable(R),
	pattern(R,N),
        !,
	call(R).
unfold_choicepoint(R,R,P,P) :-
        sp_builtin(R),
	!.
unfold_choicepoint(R,true,[(1,N)|P],P) :-
        evaluable(R), /* assume that R has a finite set of answers */
	pattern(R,N),
        !,
        call(R).
unfold_choicepoint(R,B,[(1,C)|P],P) :-
        ref_clause(R,B,C).

/* Determinate unfolding only */

prefix(true,Ns,Ns,true,_) :-
        !.
prefix(Gs,Ns,Ns1,Gs1,det) :-
        select_lit(Gs,B,Gs3,N),
        choices(B,Ch,Gs3),
        Ch=[(C,(H :- R))],
        !,
        replace_goal(Gs,N,Gs2,H,R),
        prefix(Gs2,Ns,[(N,C)|Ns1],Gs1,det).
prefix(Gs,Ns,Ns,Gs,_) :-
        no_Deadend(Gs). 

prefix1(true,Ns,Ns,true,_) :-
        !.
prefix1(Gs,Ns,Ns1,Gs1,det) :-
        select_lit(Gs,B,N),
        choices(B,Ch,Gs),
        Ch=[(C,(H :- R))],
	\+ repeat_clause(C,Ns1),
        !,
        replace_goal(Gs,N,Gs2,H,R),
        prefix1(Gs2,Ns,[(N,C)|Ns1],Gs1,det).
prefix1(Gs,Ns,Ns,Gs,_) :-
        no_Deadend(Gs). 

% repeat_clause(C,L) :-
% 	memb1((_,C),L).

repeat_clause(C,[choice|L]) :-
	memb1((_,C),L).
repeat_clause(C,[_|L]) :-
	repeat_clause(C,L).

select_lit((G,Gs),G,Gs,1).
select_lit((B,Gs),G,(B,Gs1),N1) :-
        !,
        select_lit(Gs,G,Gs1,N),
        N1 is N+1.
select_lit(G,G,true,1).


replace_goal((G,true),2,G,_,_) :- /* eliminate true at end of goal */
	!.
replace_goal((G,Gs),1,Gs1,G,R) :-
	!,
	joingoals(R,Gs,Gs1).
replace_goal((G,Gs),N,(G,Gs1),H,R) :-
	!,
	N1 is N-1,
	replace_goal(Gs,N1,Gs1,H,R).
replace_goal(G,1,R,G,R).


choices(clause(G,B),Cs,Gs) :-
	!,
	bagof((N,(clause(G,B) :- true)),[Gs]^(ref_clause(G,B,N),no_Deadend(B),
		no_Deadend(Gs)),Cs).
choices(G,Ps,Gs) :-
	sp_builtin(G),
	unfoldable(G),
	!,
	pattern(G,P),
	bagof((P,(G :- true)),[Gs]^(call(G),no_Deadend(Gs)), Ps).
choices(G,[],_) :-
	sp_builtin(G),
	!.
choices(G,[],_) :-
	imported(G),
	!.
choices(G,Ps,Gs) :-
	evaluable(G),
	!,
	pattern(G,P),
	bagof((P,(G :- true)),[Gs]^(call(G),no_Deadend(Gs)), Ps).
choices(G,Cs,Gs) :-
	bagof((N,(G :- B)),
		[Gs]^(ref_clause(G,B,N),no_Deadend(B),no_Deadend(Gs)), Cs).


