/* computes the determinate unfolding rule */
/* JPG: 10.7.90 */
/* using also regular type information */
/* JPG: 25.5.93 */

resultant(A,R) :-
        resultant(A,_,R).

resultant(G,G,R1) :-
	typed_atom(G,G1,T),
	ref_clause(G1,B,_),
	no_Deadend(B),
	types_ok(T),
	prefix(B,_,R,T),
	get_types(R,RT),
	joingoals(T,RT,U),
	unfold_body(U,U1),
	intersect_body(U1,U2),
	type_abstract(R,U2,R1).
	

typed_atom(G,G1,T) :-
	functor(G,P,N),
	functor(G1,P,N),
	G =.. [P|Ts],
	G1 =.. [P|Xs],
	type_conj(Ts,Xs,T).

type_conj([],[],true) :-
	!.
type_conj([T],[X],T1) :-
	!,
	T1 =.. [T,X].
type_conj([T|Ts],[X|Xs],(S,Ss)) :-
	S =.. [T,X],
	type_conj(Ts,Xs,Ss).
		
get_types(true,true) :-
	!.
get_types((B,Bs),R) :-
	!,
	clause(type(B),T),
	get_types(Bs,R1),
	joingoals(T,R1,R).
get_types(B,R) :-
	clause(type(B),R).

unfold_body(U,V) :-
	unfold_body(U,[],V).

intersect_body(U,V) :-
	intersect_body(U,[],_),
	intersect_body1(U,V).

intersect_body(true,_,true) :-
        !.
intersect_body((B,Bs),R,C) :-
        reoccurs(B,Bs,B1,Bs1),
        !,
        B =.. [T1,X],
        B1 =.. [T2,X],
        intersect(T1,T2,T3,P,R,0,_),
        B2 =.. [T3,X],
        intersect_body((B2,Bs1),[P|R],C).
intersect_body((B,Bs),R,(B,Bs1)) :-
        !,
        intersect_body(Bs,R,Bs1).
intersect_body(B,_,B).

intersect_body1(U,V) :-
	vars(U,Xs),
	collect_var_types(Xs,U,V).

collect_var_types([],_,true) :-
	!.
collect_var_types([X],U,V) :-
	!,
	type_occurrences(X,U,true,L),
	new_type_name(L,T),
	V =.. [T,X].
collect_var_types([X|Xs],U,(V,Vs)) :-
	type_occurrences(X,U,U1,L),
	new_type_name(L,T),
	V =.. [T,X],
	collect_var_types(Xs,U1,Vs).

type_occurrences(X,(B,Bs),U1,Ts) :-
	B =.. [T,Y],
	X == Y,
	!,
	type_occurrences(X,Bs,U1,Ts1),
	insert_ord(T,Ts1,Ts).
type_occurrences(X,(B,Bs),(B,U1),Ts) :-
	!,
	type_occurrences(X,Bs,U1,Ts).
type_occurrences(X,B,true,[T]) :-
	B =.. [T,Y],
	X == Y,
	!.
type_occurrences(_,B,B,[]).
	

new_type_name([T],T) :-
	!.
new_type_name([any|Ts],T) :-
	!,
	new_type_name(Ts,T).
new_type_name([T|Ts],S) :-
	new_type_name(Ts,S1),
	name(T,NT),
	name(S1,NS1),
	my_append(NT,[95|NS1],NS),
	name(S,NS).

insert_ord(X,L,L) :-
        memb1(X,L),
        !.
insert_ord(X,[],[X]).
insert_ord(X,[Y|Ys],[Y|Zs]) :-
        X @> Y,
        insert_ord(X,Ys,Zs).
insert_ord(X,[Y|Ys],[X,Y|Ys]) :-
        X @< Y.

type_abstract(true,_,true) :-
	!.
type_abstract((B,Bs),U,(D,Ds)) :-
	!,
	atom_abstract(B,U,D),
	type_abstract(Bs,U,Ds).
type_abstract(B,U,D) :-
	atom_abstract(B,U,D).
	
atom_abstract(A,_,A) :-
	sp_builtin(A),
	!.
atom_abstract(A,U,B) :-
	A =.. [F|As],
	clause(type(A),Ts),
	arg_type_abstract(As,Ts,U,Bs),
	B =.. [F|Bs].

arg_type_abstract([],_,_,[]).
arg_type_abstract([X|Xs],(_,Bs),U,[T|Ts]) :-
	var(X),
	arg_type_occurrence(X,U,T),
	!,
	arg_type_abstract(Xs,Bs,U,Ts).
arg_type_abstract([X],_,U,[T]) :-
	var(X),
	arg_type_occurrence(X,U,T),
	!.
arg_type_abstract([_|Xs],(B,Bs),U,[T|Ts]) :-
	functor(B,T,1),
	arg_type_abstract(Xs,Bs,U,Ts).
arg_type_abstract([_],B,_,[T]) :-
	functor(B,T,1).
	


arg_type_occurrence(X,(B,_),T) :-
        B =.. [T,Y],
        X == Y,
        !.
arg_type_occurrence(X,(_,Bs),T) :-
	arg_type_occurrence(X,Bs,T).
arg_type_occurrence(X,B,T) :-
        B =.. [T,Y],
        X == Y.
	

choicelist(G,[var]) :-
	var(G),
	!.
choicelist((G,Gs),[Ch|Chs]) :-
	!,
	choice(G,Ch),
	choicelist(Gs,Chs).
choicelist(G,[Ch]) :-
	choice(G,Ch).


prefix(true,[],true,_) :-
	!.
prefix(Gs,[(N,C)|Ns],Gs1,T) :-
	det_eval(Gs,Gs2,N,C,T),
	!,
	prefix(Gs2,Ns,Gs1,T).
prefix(Gs,[],Gs,_).

det_eval(Gs,Gs1,N,C,T) :-
	select_lit(Gs,B,Gs3,N),
	choices(B,Ch,Gs3,T),
	Ch = [(C,(H :- R))], /* determinate */
	!,
	replace_goal(Gs,N,Gs1,H,R).


select_lit((G,Gs),G,Gs,1).
select_lit((B,Gs),G,(B,Gs1),N1) :-
        !,
        select_lit(Gs,G,Gs1,N),
        N1 is N+1.
select_lit(G,G,true,1).

replace_goal((G,true),2,G,_,_) :- /* eliminate true at end of goal */
	!.
replace_goal((G,Gs),1,Gs1,G,R) :-
	!,
	joingoals(R,Gs,Gs1).
replace_goal((G,Gs),N,(G,Gs1),H,R) :-
	!,
	N1 is N-1,
	replace_goal(Gs,N1,Gs1,H,R).
replace_goal(G,1,R,G,R).

choices(clause(G,B),Cs,Gs,T) :-
	!,
	bagof((N,(clause(G,B) :- true)),[Gs]^(ref_clause(G,B,N),no_Deadend(B),
		no_Deadend(Gs),types_ok(T)),Cs).
choices(G,Ps,Gs,T) :-
	sp_builtin(G),
	unfoldable(G),
	!,
	pattern(G,P),
	bagof((P,(G :- true)),[Gs]^(call(G),no_Deadend(Gs),types_ok(T)), Ps).
choices(G,[],_,_) :-
	sp_builtin(G),
	!.
choices(G,[],_,_) :-
	imported(G),
	!.
choices(G,Ps,Gs,T) :-
	evaluable(G),
	!,
	pattern(G,P),
	bagof((P,(G :- true)),[Gs]^(sp_solve(G),no_Deadend(Gs),
			types_ok(T)), Ps).
choices(G,Cs,Gs,T) :-
	bagof((N,(G :- B)),
		[Gs]^(ref_clause(G,B,N),no_Deadend(B),no_Deadend(Gs),
			types_ok(T)), Cs).

types_ok(T) :-
	copyterm(T,T1),
	call(T1),
	!.
