% Query-answer transformation: basic version
% JPG: 27.3.92
%
% magic(F,F1,Q): magic(+,+,+)
%
%	F = source file
%	F1 = file for writing transformed program
%	Q = (atomic) query to generate the transformation
%
% e.g. ?- magic('rev.pl', transf_rev, rev(_,_)).
%
% if using with BIM, execute the query ?- bim_options.  first.
% if using with Sicstus, un-comment the last clause(s).  See comments
% at end of program.


magic(F,F1,Q) :-
	see(F),
	tell(F1),
	read(C),
	ans_clauses(C),
	seen,
	see(F),
	read(C1),
	query_clauses(C1),
	flat_name(query(Q),QQ),
	canonical( QQ ),
	displayqRClauses([(QQ :- true)]),
	nl,
	seen,
	told.

ans_clauses(X) :- 
	end_of_file(X),
	!.
ans_clauses((A :- B)) :- 
	!,
	answerlits(B,Bs),
	flat_name('_ans'(A),AA),
	flat_names(Bs,FBs),
	canonical((AA :- FBs)),
	displayqRClauses([(AA :- FBs)]),
	read(C1),
	!,
	ans_clauses(C1).
ans_clauses(A) :- 
	flat_name('_ans'(A),AA),
        canonical((AA :- true)),
        displayqRClauses([(AA :- true)]),
	read(C1),
	!,
	ans_clauses(C1).

answerlits((B,Bs),(B,Bs1)) :-
	sp_builtin(B),
	!,
	answerlits(Bs,Bs1).
answerlits((B,Bs),('_ans'(B),Bs1)) :-
	!,
	answerlits(Bs,Bs1).
answerlits(B,B) :-
	sp_builtin(B).
answerlits(B,'_ans'(B)).

flat_name(A,A) :-
	sp_builtin(A),
	!.
flat_name(A,A1) :-
	A =.. [F,As],
	As =.. [G|Args],
	name(F,Fn),
	name(G,Gn),
	concat(Gn,[95|Fn],An),
	name(H,An),
	A1 =.. [H|Args].

flat_names((B,Bs),(B1,Bs1)) :-
	!,
	flat_name(B,B1),
	flat_names(Bs,Bs1).
flat_names(B,B1) :-
	flat_name(B,B1).

query_clauses(Z) :- 
	end_of_file(Z),
	!.
query_clauses((A :- B)) :- 
	!,
	bodylitqueries(A,B),
	read(C1),
	!,
	query_clauses(C1).
query_clauses(_) :- 
	read(C1),
	!,
	query_clauses(C1).


bodylitqueries(_,true) :-
        !.
bodylitqueries(A,B) :-
        removelast(B,Bj,Bs1),
        sp_builtin(Bj),
        !,
        bodylitqueries(A,Bs1).
bodylitqueries(A,B) :-
        removelast(B,Bj,Bs1),
        !,
        answerlits(Bs1,Bs2),
        flat_names(Bs2,QBs2),
        flat_name(query(Bj),QBj),
        flat_name(query(A),QA),
        canonical((QBj :- QA,QBs2)),
        displayqRClauses([(QBj :- QA,QBs2)]),
        bodylitqueries(A,Bs1).


removelast((B,C,Bs),Bn,(B,Bs1)) :-
	!,
	removelast((C,Bs),Bn,Bs1).
removelast((B,C),C,B) :-
	!.
removelast(B,B,true).
	
concat([],Xs,Xs).
concat([X|Xs],Ys,[X|Zs]) :-
	concat(Xs,Ys,Zs).

% BIM-specific - not needed when using Sicstus

%bim_options :-
%	please(readeoffail,off),
%	please(readeofatom,end_of_file).

% end of BIM-specific

% the following to be used with Sicstus

% sp_builtin(G) :-
%         predicate_property(G,built_in).

