% Convert an atom to an RUL clause

atom_to_rul(A,Def) :-
	atom_to_rul(A,Def,0,_).

atom_to_rul(T, [proc(P/1,[(S :- B)])|Argtypes],K,K2) :-
        functor(T,F,N),
        functor(X,F,N),
        newname(K,P),
        S =.. [P,X],
        canonical(S),
        K1 is K+1,
        rul_argtypes(T,X,B,1,N,K1,K2,Argtypes).

rul_argtypes(_,_,true,_,0,K,K,[]).
rul_argtypes(T,X,B,N,N,K,K1,Argtypes) :-
        rul_argtype(T,X,B,N,K,K1,Argtypes).
rul_argtypes(T,X,(B,Bs),M,N,K,K2,Argtypes) :-
        M < N,
        rul_argtype(T,X,B,M,K,K1,Argtype),
        M1 is M+1,
        rul_argtypes(T,X,Bs,M1,N,K1,K2,Argtypes1),
        app(Argtype,Argtypes1,Argtypes).

rul_argtype(T,X,B,M,K,K,[]) :-
        arg(M,T,Xm),
        var(Xm),
        !,
        arg(M,X,Ym),
        B =.. [any,Ym].
rul_argtype(T,X,B,M,K,K1,Types) :-
        arg(M,T,Xm),
        arg(M,X,Ym),
        rul_termtype(Xm,Types,K,K1),
        newname(K,P),
        B =.. [P,Ym].

rul_termtype(T, [proc(P/1,[(S :- B)])|Argtypes],K,K2) :-
        functor(T,F,N),
        functor(X,F,N),
        newname(K,P),
        S =.. [P,X],
        canonical(S),
        K1 is K+1,
        rul_argtypes(T,X,B,1,N,K1,K2,Argtypes).

% upper bound of two atoms

atom2_ub(A1,A2,UB,K0,K1) :-
	copyterm(A1,CA1),
	copyterm(A2,CA2),
	canonical(CA1),
	canonical(CA2),
	lub(catom(CA1,(true,[])),catom(CA2,(true,[])),UB,K0,K1).

% Introduce recursive definitions

widen_rul(RP,RP4,K,K3) :-
        head_functors(RP,Fs),
        same_head_functors(T1,T2,Fs),
        depends(T1,T2,TPs,RP),
        subtype(T2,T1,RP),
        !,
        recursive(T1,T2,TPs,RP,RP1,K,K1),
	rename_all_preds(RP1,RP2,_,K1,K2),
        toptype(RP2,T),
        rdefs(T,RP3,RP2),
        widen_rul(RP3,RP4,K2,K3).
widen_rul(R,R,K,K).


% Compute upper bound of catom(A,Defs) and widen. 


lub(catom(A1,(U1,Def1)), catom(A2,(U2,Def2)), catom(A3,(U3,Def3)), K1,K2) :-
	melt((A1,U1),(CA1,V1)),
	melt((A2,U2),(CA2,V2)),
	msg1(CA1,CA2,A3,L),
	build_rul_args(L,V1,V2,Def1,Def2,U3,Def3,K1,K2),
	canonical((A3,U3)).
	
build_rul_args([],_,_,_,_,true,[],K,K) :-
	!.
build_rul_args([subst(T1,T2,X)],U1,U2,Def1,Def2,B,Def4,K0,K1) :-
	!,
	build_rul_arg(T1,T2,X,U1,U2,Def1,Def2,B,Def4,K0,K1).
build_rul_args([subst(T1,T2,X)|L],U1,U2,Def1,Def2,(B,Bs),Def45,K0,K2) :-
	build_rul_arg(T1,T2,X,U1,U2,Def1,Def2,B,Def4,K0,K1),
	build_rul_args(L,U1,U2,Def1,Def2,Bs,Def5,K1,K2),
	app(Def4,Def5,Def45).
	
build_rul_arg(T1,T2,X,U1,U2,Def1,Def2,B,Def5,K0,K4) :-
	termtype1(T1,U1,P1,D1,K0,K1),
	termtype1(T2,U2,P2,D2,K1,K2),
	upperbound(P1,P2,D1,D2,Def1,Def2,X,B,Def5,K2,K4).


upperbound(any,_,_,_,_,_,X,any(X),[],K,K) :-
	!.
upperbound(_,any,_,_,_,_,X,any(X),[],K,K) :-
	!.
upperbound(P1,P2,D1,D2,Def1,Def2,X,B,Def5,K2,K4) :-
	app(D1,Def1,DDef1),
	app(D2,Def2,DDef2),
	app(DDef1,DDef2,DDef),
	ub0(P1,P2,P,PDef,DDef,K2,K3),
	app(PDef,DDef,PDDef),
	rdefs0(P,Def4,PDDef),
	widen_rul(Def4,Def5,K3,K4),
	toptype(Def5,T),
	B =.. [T,X].

rdefs0(any,[],_) :-
	!.
rdefs0(T,R,D) :-
	rdefs(T,R,D).
	
ub0(any,_,any,[],_,K,K) :-
	!.
ub0(_,any,any,[],_,K,K) :-
	!.
ub0(P1,P2,P,PDef,DDef,K2,K3) :-
	ub1(P1,P2,P,PDef,DDef,K2,K3).
	
termtype1(T1,U1,P,[],K,K) :-
	var(T1),
	!,
	giventype(T1,U1,P).
termtype1(T1,U1,P,D,K0,K1) :-
	headtype(T1,U1,P,D,K0,K1).
	
rulatom_equal(catom(A1,(U1,R1)),catom(A2,(U2,R2))) :-
	A1 == A2,
	app(R1,R2,R3),
	pairwise_rul_eq(U1,U2,R3).
	
pairwise_rul_eq(true,true,_) :-
	!.
pairwise_rul_eq((B1,U1),(B2,U2),R3) :-
	!,
	B1 =.. [T1,_],
	B2 =.. [T2,_],
	subtype(T1,T2,R3),
	subtype(T2,T1,R3),
	pairwise_rul_eq(U1,U2,R3).
pairwise_rul_eq(B1,B2,R3) :-
	B1 =.. [T1,_],
	B2 =.. [T2,_],
	subtype(T1,T2,R3),
	subtype(T2,T1,R3).
	
% Remove any determinate arguments in Def1 and propagate the terms to A

detdefs(A,[proc(T/1,[(H :- B1)])|Def1],Def2) :-
	A =.. [_|Args],
	detelim(Args,B1,B2,Def1),
	rdefs(T,Def2,[proc(T/1,[(H :- B2)])|Def1]).
	
detelim([],true,true,_).
detelim([X],B1,any(Xn),Def1) :-
	B1 =.. [T,Xn],
	S =.. [T,X],
	detsolve(S,Def1),
	!.
detelim([_],B1,B1,_) :-
	!.
detelim([X1,X2|Xs],(B1,Bs),(any(Xn),Bs1),Def1) :-
	B1 =.. [T,Xn],
	S =.. [T,X1],
	detsolve(S,Def1),
	!,
	detelim([X2|Xs],Bs,Bs1,Def1).
detelim([_,X2|Xs],(B1,Bs),(B1,Bs1),Def1) :-
	detelim([X2|Xs],Bs,Bs1,Def1).
	
