/* read in a program, for Sicstus version of SP system */
/* use of this procedure avoids the need to 
   declare each predicate as dynamic */


sp_consult(F) :-
	sp_ttynl,
	nofileerrors,
	see(F),
	fileerrors,
	abolish_all_existing,
	read(C),
	read_clauses(C,Ps,1),
	seen,
	sp_ttynl,
	write(user,'Finished reading '),write(user,F),
	sp_ttynl,
	assert(program_preds(Ps)),
	strong_connected_components(Ps,SCCs),
	put_sccs(SCCs),
	!.
sp_consult(F) :-
	name(F,FN),
	name('.pl',S),
	file_suffix(FN,S,FS),
	name(F1,FS),
	nofileerrors,
	see(F1),
	fileerrors,
	abolish_all_existing,
	read(C),
	read_clauses(C,Ps,1),
	seen,
	sp_ttynl,
	write(user,'Finished reading '),write(user,F1),
	sp_ttynl,
	assert(program_preds(Ps)),
	strong_connected_components(Ps,SCCs),
	put_sccs(SCCs),
	!.
sp_consult(F) :-
	seen,
	assert(program_preds([])),
	write(user,'Cannot find file '), write(user,F),
	write(user,', or problems while reading.'), sp_ttynl,sp_ttynl.

read_clauses(end_of_file,Ps,N) :- 
	!,
	write(user,'Number of clauses: '), write(user,N),
	sp_ttynl,
	close_list(Ps).
read_clauses((:- entry(G,_)),Ps,N) :-
        !,
        assert(directive(entry(G))),
        read(C1),
        read_clauses(C1,Ps,N).
%read_clauses((:- import(P/N)),Ps,K) :-
%        !,
%        assert(directive(import(P/N))),
%	functor(H,P,N),
%	assert(imported(H)),
%        read(C1),
%        read_clauses(C1,Ps,K).
%read_clauses((:- import(from(P/N,M))),Ps,K) :-
%        !,
%        assert(directive(import(from(P/N,M)))),
%	functor(H,P,N),
%	assert(imported(H)),
%        read(C1),
%        read_clauses(C1,Ps,K).
read_clauses((:- dynamic(P/N)),Ps,K) :-
        !,
        assert(directive(dynamic(P/N))),
	functor(H,P,N),
	assert(ub_clause(imported(H),true)),
        read(C1),
        read_clauses(C1,Ps,K).
read_clauses((:- Dir),Ps,N) :-
        !,
        call(Dir),
	assert(directive(Dir)),
        read(C1),
        !,
        read_clauses(C1,Ps,N).
read_clauses(C,Ps,N) :- 
	get_pred_name(C,Pred,Bodypreds),
	each_memb1([Pred|Bodypreds],Ps),
	store_clause(C),
	read(C1),
	!,
	N1 is N+1,
	read_clauses(C1,Ps,N1).

get_pred_name((H :- B),P/N,BPs) :-
	!,
	functor(H,P,N),
	body_preds(B,BPs).
get_pred_name(H ,P/N,[]) :-
	functor(H,P,N).

body_preds(true,[]) :-
	!.
body_preds((\+ B,Bs),Ps) :-
	!,
	body_preds((B,Bs),Ps).
body_preds((B,Bs),Ps) :-
	sp_builtin(B),
	!,
	body_preds(Bs,Ps).
body_preds((B,Bs),[P/N|Ps]) :-
	!,
	functor(B,P,N),
	body_preds(Bs,Ps).
body_preds(\+ B,Ps) :-
	!,
	body_preds(B,Ps).
body_preds(B,[]) :-
	sp_builtin(B),
	!.
body_preds(B,[P/N]) :-
	functor(B,P,N).

each_memb1([],_).
each_memb1([P|Ps],S) :-
	memb1(P,S),
	each_memb1(Ps,S).

file_suffix([],X,X).
file_suffix([X|Xs],Ys,[X|Zs]) :-
	file_suffix(Xs,Ys,Zs).

close_list([]) :-
	!.
close_list([_|X]) :-
	close_list(X).

abolish_all_existing :-
	!,
	write(user,'Clearing workspace....'),
	abolish(ub_clause/2),
	abolish(directive/1),
	assert(directive(dummy)),
	abolish(program_preds/1),
	abolish(program_pred/2).
abolish_all_existing.

put_sccs([]).
put_sccs([(R,C)|Cs]) :-
	put_scc(R,C),
	put_sccs(Cs).

put_scc(_,[]).
put_scc(R,[P|Ps]) :-
	assert(program_pred(P,R)),
	put_scc(R,Ps).

store_clause((H :- B)) :-
        !,
        assert(ub_clause(H,B)).
store_clause(H) :-
        assert(ub_clause(H,true)).

