/* sp.pl */
/* Construct specialised program */
/* 
	1.  Flow analysis performed by abint/3
	2.  Renaming of predicates performed by rename_preds/4
	3.  Construction of new procedures by make_clauses/5
	4.  Writing out to file by write_to_file/1.

*/
/* This core procedure can be adapted to perform
   various kinds of specialisation.  Different abstraction
   operations, and unfolding transformations can be defined.
   The following procedures define variants of the program:

	resultant/2
	more_specific/1
	cleanup/3

 */

/* John Gallagher: 31.7.90 */
/*  revised 31.8.90 */

%:- setdebug.

%:- dynamic directive/1.

sp(G) :-
	vars(G,Xs),
	anybody(Xs,U),
	flow_analyse(catom(G,(U,[])),Pfg,Pathtable,0,K),
	rename_preds(catom(G,(U,[])),Pfg,Pathtable,R),
	!,
	make_clauses(catom(G,(U,[])),R,Pfg,Pathtable,Cls,K),
	!,
	write_to_file(Cls),
        !.

anybody([],true) :-
	!.
anybody([X],any(X)) :-
	!.
anybody([X|Xs],(any(X),As)) :-
	anybody(Xs,As).

flow_analyse(G,Pfg,Pathtable,K0,K1) :-
        nl,
        write('flow analysis---'),sp_ttyflush,
        abint(G,Pfg,Pathtable,K0,K1).

rename_preds(G,Pfg,Pathtable,R) :-
	rep(G,Initrep,Pathtable,Pfg),
	G = catom(G1,_),
	functor(G1,P,_),
	R = [name(Initrep,0,P)|_],
	renames(Pfg,R).

make_clauses(G,R,Pfg,Pathtable,Cls,K0) :-
        nl,write('constructing specialised program---'),sp_ttyflush,
        tell(tempwrite),procedures(R,R,Cls1,Pathtable,Pfg,K0,_),told,
        nl,nl,write('cleaning up---'),sp_ttyflush,
        cleanup(G,Cls1,Cls).

write_to_file(Cls) :-
        nl,write('Enter filename: '),
        read(F),
        sp_filewrite(F,Cls).

/* Rename each atom in the environment */

renames([],R) :-
	!,
	list1(R).
renames([element(_,As,_,_)|Cs],R) :-
	most_general(As,C),
	renamepred(C,R),
	renames(Cs,R).
	
renamepred(P,S) :-
	renamepred(P,S,0).

renamepred(catom(A,Cs),S,M) :-
	var(S),
	!,
	N is M + 1,
	A =.. [P|_],
	name(P,Plist),
	name(N,Nlist),
	my_append(Plist,[95|Nlist],PNlist), /* 95 = underscore */
	name(Pn,PNlist),
	S = [name(catom(A,Cs),N,Pn)|_],
	!.
renamepred(P,[name(P1,_,_)|_],_) :-
	rulatom_equal(P,P1),
	!.
renamepred(catom(P,Cs),[name(catom(Q,_),M,_)|S],_) :-
	P =.. [A|_],
	Q =.. [A|_],
	!,
	renamepred(catom(P,Cs),S,M).
renamepred(P,[_|S],K) :-
	renamepred(P,S,K).

/*  Generate procedures for the calls  */

procedures([],_,[],_,_,K,K) :-
	!.
procedures([name(A,N,Pn)|Rs],R,[Cls|Prs],Env,Pfg,K0,K2) :-
      procedure(A,Env,R,N,Pn,Cls,Pfg,K0,K1),
	!,
      procedures(Rs,R,Prs,Env,Pfg,K1,K2).

procedure(catom(A,(U,C)),Env,Rs,K,Pn,Cls,Pfg,K0,K1) :- 
	!,
	nl,nl,
	write(A),
	write(' :- '),
	nl,
	write('      '),
	write(U),
	write('.'), 
	nl,
	displayqTypes(C),
	retrieve_resultants(catom(A,(U,C)),Env,Cls1),
	rename_bodies(A,Cls1,Env,Pfg,Rs,Cls,K,Pn,K0,K1).

/*  Retrieve the set of resultants for a given atom */

retrieve_resultants(A,Pathtable,Rs) :-
	get_path(A,Pathtable,_,Rs,_,_).


rename_bodies(_,[],_,_,_,[],_,_,K,K) :-
	!.
rename_bodies(_,[(Resultant,UDefs)|Cls1],Env,Pfg,Rs,[(H :- R2)|Cls],0,_,K0,K2) :-
	melt(Resultant,(H1 :- R1,U1)),
	H1 =.. [P,_|Args],
	H =.. [P|Args],
	make_body(R1,U1,UDefs,Env,Pfg,Rs,R2,K0,K1),
	!,
	rename_bodies(_,Cls1,Env,Pfg,Rs,Cls,0,_,K1,K2).
rename_bodies(A,[(Resultant,UDefs)|Cls1],Env,Pfg,Rs,[(H :- R2)|Cls],J,Pn,K0,K2) :-
	J > 0,
	melt(A,H1),
	H1 =.. [_,_|Args],
	vars(Args,Xs),
	melt(Resultant,(H1 :- R1,U1)),
	H =.. [Pn|Xs],
	make_body(R1,U1,UDefs,Env,Pfg,Rs,R2,K0,K1),
	!,
	rename_bodies(A,Cls1,Env,Pfg,Rs,Cls,J,Pn,K1,K2).
rename_bodies(A,[_|Cls1],Env,Pfg,Rs,Cls,K,Pn,K0,K1) :- 	/* couldn't make body */
	rename_bodies(A,Cls1,Env,Pfg,Rs,Cls,K,Pn,K0,K1). /* because it fails   */

make_body(B,_,_,_,_,_,B,K,K) :- 
	var(B),
	!.
make_body((B,Bs),U,UDefs,Env,Pfg,Rs,(A,As),K0,K2) :- 
	!,
	buildNewGoal(B,U,UDefs,U1,UDefs1,Env,Pfg,Rs,A,K0,K1),
	make_body(Bs,U1,UDefs1,Env,Pfg,Rs,As,K1,K2).
%make_body(B,U,UDefs,Env,Pfg,Rs,B1,K0,K1) :- 
%	prolog_meta(B,B1,P,P1),
%	!,
%	make_body(P,U,UDefs,Env,Pfg,Rs,P1,K0,K1).
make_body(B,U,UDefs,Env,Pfg,Rs,A,K0,K1) :-
	buildNewGoal(B,U,UDefs,_,_,Env,Pfg,Rs,A,K0,K1).

buildNewGoal(G,U,Udefs,U,Udefs,_,_,_,G,K,K) :-
	sp_builtin(G),
	!.
buildNewGoal(true,U,Udefs,U,Udefs,_,_,_,true,K,K) :-
	!,
	soluble(U,Udefs,_,_,K,_).
buildNewGoal(G,U,Udefs,U,Udefs,_,_,_,G,K,K) :-
	imported(G),
	!.
buildNewGoal(B,U,UDefs,U2,UDefs2,Env,Pfg,Rs,B1,K0,K1) :-
	constrained_atom(B,U,UDefs,[Call]),
	rep(Call,R,Env,Pfg),
	renameGoal(R,B,B1,Rs),
	insert_ans(B,Call,U,UDefs,U2,UDefs2,Env,Pfg,K0,K1).


insert_ans(B,Call,U1,UDefs1,U3,UDefs3,Env,Pfg,K0,K1) :-
	find_path(Call,Env,Path),
	find_ans(Path,B,Pfg,U,UDefs),
	joingoals(U,U1,U2),
	app(UDefs,UDefs1,UDefs2),
	soluble(U2,UDefs2,U3,S,K0,K1),
      app(S,UDefs2,UDefs3).


%renameGoal(fail,_,fail,_) :-
%	!.
renameGoal(catom(R,C),B,B1,Rs) :-
%	memb2(name(catom(R,C),0,_),Rs),
	memb2(name(CA,0,_),Rs),
	rulatom_equal(catom(R,C),CA),
	!,
	B =.. [P,_Trace|Args],
	B1 =.. [P|Args].
renameGoal(catom(R,C),B,A,Rs) :-
%	memb2(name(catom(R,C),_,Pn),Rs),
	memb2(name(CA,_,Pn),Rs),
	rulatom_equal(catom(R,C),CA),
	melt(R,Mr),
	Mr =.. [_,_Trace|Args],
	vars(Args,Vs),
	A =.. [Pn|Vs],
	B = Mr,
	!.
renameGoal(_,B,_,_) :-
	nl,
	write(B),
	write('cannot be renamed.......'),
	nl,nl,
	fail.

 
most_general([A|As],A) :-
	var(As),
	!.
most_general([_|As],A) :-
	most_general(As,A).

rep(A,B,Pathtable,Pfg) :-
%	get_path(A,Pathtable,P,_,_,_),   %This should be 'find_path'
	find_path(A,Pathtable,P),
	get_reps(P,Pfg,Bs),
	most_general(Bs,B),
	!.
rep(_,fail,_,_) :-
	nl,
	write('Failed atom detected:  replaced by ''fail'''),
	nl.

get_reps(P,[element(P1,Bs,_,_)|_],Bs) :-
	P == P1,
	!.
get_reps(P,[_|Es],Bs) :-
	get_reps(P,Es,Bs).

sp_filewrite(tty,Cls) :-
        !,
	nl,write('Specialised Program:'),nl,nl,
        displayqProcs(Cls),
	nl,nl.
sp_filewrite(F,Cls) :-
        tell(F),
	write_directives,
        displayqProcs(Cls),
        told.

/* write the program to current output file  */

displayqProcs([P|Ps]) :-
        displayqClauses(P),
        displayqProcs(Ps).
displayqProcs([]).

write_directives :-
	bagof(X,(directive(X),X \== dummy,write_directive(X)),_),
	!.
write_directives.

write_directive(X) :-
	canonical(X),
	write(':- '),
	write(X),
	write('.'),
	nl.

imported(X) :- ub_clause(imported(_,X),true).

cleanup(_,P,P).

