:- dynamic evaluable/1.


choice((G,_),Cs) :-
	!,
	choice1(G,Cs).
choice(G,Cs) :-
	choice1(G,Cs).

choice1(clause(G,B),Cs) :-
	!,
	bagof(N,[G,B]^(ref_clause(G,B,N),no_Deadend(B)),Cs).
choice1(call(G),P) :-
	!,
	pattern(G,P).
choice1(G,P) :-
	imported(G),
	!,
	functor(G,P,_).
choice1(G,P) :-
	sp_builtin(G),
	!,
	pattern(G,P).
choice1(G,Cs) :-
	bagof(N,[G,B]^(ref_clause(G,B,N),no_Deadend(B)),Cs).


pattern(G,P) :-
	G =.. [P],
	!.
pattern(G,[P|P1]) :-
	G =.. [P,A1],
	classify(A1,P1).
pattern(G,[P,P1|P2]) :-
	G =.. [P,A1,A2|_],
	classify(A1,P1),
	classify(A2,P2).

classify(A1,A1) :-
	atom(A1),
	!.
classify(A1,var) :-
	var(A1),
	!.
classify(A1,num) :-
	number(A1),
	!.
classify(A1,F) :-
	nonvar(A1),
	!,
	A1 =.. [F|_].
classify(_,x).

solve_choice((G,Gs),Gs) :-
	!,
	clause_heads(G).
solve_choice(G,true) :-
	clause_heads(G).

clause_heads(clause(A,_)) :-
	!,
	bagof(A,[B]^
		ub_clause(A,B), Hs),
	setMsg(Hs,A).
clause_heads(G) :-
	sp_builtin(G),
	!.
clause_heads(G) :-
	imported(G),
	!.
clause_heads(A) :-
	bagof(A,[B]^
		ub_clause(A,B), Hs),
	setMsg(Hs,A).


unfold_choice((R,Rs),Bs) :-
	!,
	unfold_choice(R,R1),
	unfold_choice(Rs,Rs1),
	joingoals(R1,Rs1,Bs).
unfold_choice(true,true) :-
	!.
unfold_choice(clause(H,B),true) :-
	nonvar(H),
	!,
	ub_clause(H,B).
unfold_choice(R,R) :-
	imported(R),
	!.
unfold_choice(R,R) :-
	sp_builtin(R),
	!.
unfold_choice(R,B) :-
	unf(R),
	!,
	ub_clause(R,B).
/*
unfold_choice(R,true) :-
	all_unit_clauses(R),
	!,
	ub_clause(R,true).
*/
unfold_choice(R,true) :-
	evaluable(R), 
	!,
	sp_solve(R).
unfold_choice(R,R).

all_unit_clauses(R) :-
	\+ (ub_clause(R,B),B \== true).

%evaluable(member(_,Y)) :-
%	completeList(Y).
%evaluable(append(_,_,Y)) :-
%	completeList(Y).
%evaluable(neg(P)) :-
%	ground(P).

/*
more_specific(B) :-
	var(B),
	!.
more_specific((B,Bs)) :-
	!,
	more_specific(B),
	more_specific(Bs).
more_specific(B) :-
        solve_choice(B,_).

choicelist(G,[var]) :-
	var(G),
	!.
choicelist((G,Gs),[Ch|Chs]) :-
	!,
	choice(G,Ch),
	choicelist(Gs,Chs).
choicelist(G,[Ch]) :-
	choice(G,Ch).

*/

no_Deadend(B) :-
	bagof(B, (no_Deadend1(B),!),_).

no_Deadend1(G) :-
	var(G),
	!.
no_Deadend1((G,Gs)) :-
	!,
	no_Deadend1(G),
	no_Deadend1(Gs).
no_Deadend1(G) :-
	imported(G),
	!.
no_Deadend1(G) :-
	sp_builtin(G),
	unfoldable(G),
	!,
	sp_solve(G).
no_Deadend1(G) :-
	sp_builtin(G),
	!.
/*
no_Deadend1(G) :-
	evaluable(G),
	!,
	sp_solve(G).
*/
no_Deadend1(G) :-
	ub_clause(G,_).


sp_solve(true) :-
	!.
sp_solve((G,Gs)) :-
	!,
	sp_solve(G),
	sp_solve(Gs).
sp_solve(\+ G) :-
	!,
	\+ sp_solve(G).
sp_solve(G) :-
	sp_builtin(G),
	!,
	call(G).
sp_solve(G) :-
	ub_clause(G,B),
	sp_solve(B).
%sp_solve(G) :-
%	G =.. [P|Xs],
%	G1 =.. [P,_|Xs],
%	ub_clause(G1,B),
%	sp_solve(B).

