/* Utility predicates for abstract partial evaluation */


bagof1(X,P,S) :-
	bagof(X,P,S1),
	!,
	S=S1.
bagof1(_,_,[]).

/* msg(A,B,C) : C is the least common anti-instance 
   (generalisation) of A and B.  
   Algorithm from Reynolds, Machine Intelligence 5 */ 

msg(T1,T2,T3) :-
	msg1(T1,T2,T3,_).

msg1(T1,T2,T3,S) :-
	msgDiffer(T1,T2),
	msgOccur(subst(T1,T2,T3),S),
	!.
msg1(T1,T2,T3,S) :-
	nonvar(T1),
	nonvar(T2),
	!,
	T1 =.. [F|Args1],
	T2 =.. [F|Args2],
	msgArgs(Args1,Args2,Args3,S),
	T3 =.. [F|Args3].
msg1(T1,_,T1,_).

msgArgs([X|Xs],[Y|Ys],[Z|Zs],S) :-
	msg1(X,Y,Z,S),
	msgArgs(Xs,Ys,Zs,S).
msgArgs([],[],[],_).

msgDiffer(X,Y) :-
	var(X),
	!,
	X \== Y.
msgDiffer(X,Y) :-
	var(Y),
	!,
	X \== Y.
msgDiffer(X,Y) :-
	functor(X,F,N),
	functor(Y,G,M),
	msgDiffTerm(F,G,N,M).

msgDiffTerm(F,G,_,_) :-
	F \== G,
	!.
msgDiffTerm(_,_,N,M) :-
	N =\= M.

msgOccur(subst(X,Y,Z),[U|_]) :-
	var(U),
	U = subst(X,Y,Z),
	!.
msgOccur(subst(X,Y,Z),[subst(U,V,Z)|_]) :-
	X == U,
	Y == V,
	!.
msgOccur(subst(X,Y,Z),[_|S]) :-
	msgOccur(subst(X,Y,Z),S).

/* setMsg(Xs,A) : A is the msg of a set of terms Xs */

setMsg([X],X) :-
	!.
setMsg([X,Y|Xs],W) :-
	msg(X,Y,V),
	setMsg([V|Xs],W).


memb1(X,[X|_]) :-
	!.
memb1(X,[_|Xs]) :-
	memb1(X,Xs).

memb3(X,[X1|_]) :-
	X == X1,
	!.
memb3(X,[_|Xs]) :-
	memb3(X,Xs).

canonical_list([A|As],[C|Cs]) :-
	copyterm(A,C),
	canonical(C),
	canonical_list(As,Cs).
canonical_list([],[]).


/* unfoldability conditions for Sicstus builtins */

unfoldable(clause(X,_)) :-
	nonvar(X).
unfoldable(predicate_property(X,_)) :-
	nonvar(X).
unfoldable(_ is Y) :-
	arithexpr(Y).
unfoldable(_ = _).
unfoldable('C'(_,_,_)).
unfoldable(true).
% unfoldable('!').
unfoldable(X =.. _) :-
	nonvar(X),
	!.
unfoldable(_ =.. [F|Args]) :-
	atom(F),
	completeList(Args),
	!.
unfoldable(_ =.. [F]) :-
	number(F).
unfoldable(X < Y) :-
	arithexpr(X),
	arithexpr(Y).
unfoldable(X > Y) :-
	arithexpr(X),
	arithexpr(Y).
unfoldable(X =< Y) :-
	arithexpr(X),
	arithexpr(Y).
unfoldable(X >= Y) :-
	arithexpr(X),
	arithexpr(Y).
unfoldable(X @< Y) :-
	ground(X),
	ground(Y).
unfoldable(X @> Y) :-
	ground(X),
	ground(Y).
unfoldable(X @=< Y) :-
	ground(X),
	ground(Y).
unfoldable(X @>= Y) :-
	ground(X),
	ground(Y).
unfoldable(X == Y) :-
	nonvar(X),
	nonvar(Y),
	X =.. [F|_],
	Y =.. [G|_],
	F \== G,
	!.
unfoldable(X == Y) :-
	ground(X),
	ground(Y).
unfoldable(X \== Y) :-
	X == Y,
	!.
unfoldable(X \== Y) :-
	ground(X),
	ground(Y),
	!.
unfoldable(X \== Y) :-
	nonvar(X),
	nonvar(Y),
	X =.. [F|_],
	Y =.. [G|_],
	F \== G.
unfoldable(X =\= Y) :-
	arithexpr(X),
	arithexpr(Y),
	X1 is X,
	Y1 is Y,
	X1 == Y1,
	!.
unfoldable(X =\= Y) :-
	ground(X),
	ground(Y),
	!.
unfoldable(var(X)) :-
	nonvar(X).
unfoldable(nonvar(X)) :-
	nonvar(X).
unfoldable(real(X)) :-
	nonvar(X).
unfoldable(integer(X)) :-
	nonvar(X).
unfoldable(number(X)) :-
	nonvar(X).
unfoldable(atomic(X)) :-
	nonvar(X).
unfoldable(atom(X)) :-
	nonvar(X).
unfoldable(ground(X)) :-
	ground(X).
unfoldable(not(X)) :-
	ground(X).
unfoldable(\+(X)) :-
	sp_builtin(X),
	ground(X).
unfoldable(\+(X)) :-
	sp_builtin(X),
	!,
	unfoldable(X),
	\+ X.
unfoldable(\+(X)) :-
	nonvar(X),
	X =.. [_,_|Args],
	ground(Args).
unfoldable(name(X,_)) :-
	atomic(X),
	!.
unfoldable(name(_,X)) :-
	intlist(X).
unfoldable(?=(X,Y)) :-
	nonvar(X),
	nonvar(Y).


/*
unfoldable(\+(X)) :-
	\+ call(X).
*/
unfoldable(fail).
unfoldable(functor(X,_,_)) :-
	nonvar(X).
unfoldable(functor(_,Y,Z)) :-
	integer(Z),
	atomic(Y).
unfoldable(arg(X,Y,_)) :-
	integer(X),
	nonvar(Y).
unfoldable(call(X)) :-
	unfoldable(X).
unfoldable(length(X,_)) :-
	completeList(X).
unfoldable(length(_,Y)) :-
	integer(Y).

completeList(X) :-
	X == [].
completeList(Xs) :-
	nonvar(Xs),
	Xs = [_|Xs1],
	completeList(Xs1).

intlist(X) :-
	X == [].
intlist(Xs) :-
	nonvar(Xs),
	Xs = [X|Xs1],
	integer(X),
	intlist(Xs1).


memb2(X,[X|_]).
memb2(X,[_|Xs]) :-
	memb2(X,Xs).

copyterm(T1,T2) :-
	bagof(X,T1^(X = T1),[T2]).

canonical(T) :-
	canonical(1,T).

canonical(N,T) :-
	name('X',[A]),
	mynumbervars(T,A,N,_).

mynumbervars(X,Ch,N,N1) :-
	var(X),
	!,
	N1 is N+1,
	name(N,Nlist),
	name(X,[Ch|Nlist]).
mynumbervars(X,_,N,N) :-
	atomic(X),
	!.
/*
mynumbervars(T,Ch,N,N1) :-
	functor(T,F,K),
	mynumbervarargs(1,K,T,Ch,N,N1).

mynumbervarargs(I,K,_,_,N,N) :-
	I > K,
	!.
mynumbervarargs(I,K,T,Ch,N,N2) :-
	arg(I,T,Ti),
	mynumbervars(Ti,Ch,N,N1),
	I1 is I+1,
	mynumbervarargs(I1,K,T,Ch,N1,N2).
*/

mynumbervars(T,Ch,N,N1) :-
%	nonvar(T),
%	!,
	T =.. [_|Args],
	mynumbervarlist(Args,Ch,N,N1).

mynumbervarlist([],_,N,N).
mynumbervarlist([T|Ts],Ch,N,N2) :-
	mynumbervars(T,Ch,N,N1),
	mynumbervarlist(Ts,Ch,N1,N2).
	
/* Carl Seghers/ Frank Defoort's version of melt 
%%%%%%% change: faster melt (if lots of variables involved)  %%%%%%%%

melt(X,Y) :-
	functor(S,assocs,51),
	melt1(X,Y,S).

melt1([X|Xs],[Y|Ys],S) :-
	!,
	melt1(X,Y,S),
	melt1(Xs,Ys,S).
melt1(X,Y,S) :-
	variable(X,N),
	!,
	assoc(N,Y,S).
melt1(X,X,_) :-
	atomic(X),
	!.
melt1(X,Y,S) :-
	X =.. [Z|Zs],
	melt1(Z,Z1,S),
	melt1(Zs,Zs1,S),
	Y =.. [Z1|Zs1].

assoc(N,Y,S) :-
	N < 51,!,
	arg(N,S,Y).
assoc(N,Y,S) :-
	arg(51,S,NewS),
	functor(NewS,assocs,51),
	N1 is N-50,
	assoc(N1,Y,NewS).

*/

melt(X,Y) :-
	melt1(X,Y,_Assoclist),
	!.

melt1(X,Y,S) :-
	variable(X),
	!,
	assoc(X,Y,S).
melt1(X,X,_) :-
	atomic(X),
	!.
melt1(X,Y,S) :-
	functor(X,F,N),
	functor(Y,F,N),
	meltargs(1,N,X,Y,S).

meltargs(I,N,_,_,_) :-
	I > N,
	!.
meltargs(I,N,X,Y,S) :-
	arg(I,X,Xi),
	melt1(Xi,Yi,S),
	arg(I,Y,Yi),
	I1 is I+1,
	meltargs(I1,N,X,Y,S).

/*
melt1([X|Xs],[Y|Ys],S) :-
	!,
	melt1(X,Y,S),
	melt1(Xs,Ys,S).
melt1(X,Y,S) :-
	X =.. [Z|Zs],
	melt1(Z,Z1,S),
	melt1(Zs,Zs1,S),
	Y =.. [Z1|Zs1].
*/

assoc(X,Y,[assoc(X,Y)|_]) :-
	!.
assoc(X,Y,[_|S]) :-
	assoc(X,Y,S).

instanceOf(X,Y) :-
	melt(X,Mx),
	copyterm(Y,Mx),
	canonical(Mx),
	X == Mx.
	
writeByLine([]) :-
	!.
writeByLine([X|Zs]) :-
	write(X),
	nl,
	writeByLine(Zs).

/*
ground(T) :-
	nonvar(T),
	atomic(T),
	!.
ground(T) :-
	nonvar(T),
	T =.. [_|Args],
	groundlist(Args).

groundlist([]).
groundlist([T|Ts]) :-
	ground(T),
	groundlist(Ts).
*/


arithexprlist([]).
arithexprlist([T|Ts]) :-
	arithexpr(T),
	arithexprlist(Ts).

arithexpr(T) :-
        nonvar(T),
        number(T),
        !.
arithexpr(T) :-
        nonvar(T),
        T =.. [Op|Args],
        arithop(Op),
        arithexprlist(Args).

arithop(+).
arithop(-).
arithop(*).
arithop(+).
arithop(//).
arithop(mod).
arithop(integer).
arithop(float).
arithop(/\).
arithop(\/).
arithop(^).
arithop(>>).
arithop(<<).

joingoals(true,Xs,Xs) :-
	!.
joingoals(Xs,true,Xs) :-
	!.
joingoals((true,Xs),Ys,Zs) :-
	!,
	joingoals(Xs,Ys,Zs).
joingoals((Xs,true),Ys,Zs) :-
	!,
	joingoals(Xs,Ys,Zs).
joingoals((X,Xs),Ys,(X,Zs)) :-
	!,
	joingoals(Xs,Ys,Zs).
joingoals(X,Xs,(X,Xs)) :-
	X =.. [F|_],
	F \== ','.

 
my_append([],Ys,Ys).
my_append([X|Xs],Ys,[X|Zs]) :-
	my_append(Xs,Ys,Zs).

list1([]) :-
	!.
list1([_|Xs]) :-
	list1(Xs).

vars(T,Vs) :-
        vars(T,[],Vs).

vars(X,Vs,Vs1) :-
        var(X),
        !,
        insertvar(X,Vs,Vs1).
vars(X,Vs,Vs) :-
	atomic(X),
	!.
/*
vars(X,Vs,Vs1) :-
	functor(X,F,N),
	argvars(1,N,X,Vs,Vs1).

argvars(I,N,_,Vs,Vs) :-
	I > N,
	!.
argvars(I,N,X,Vs,Vs2) :-
	arg(I,X,Xi),
	functor(X,F,N),
	vars(X1,Vs,Vs1),
	I1 is I+1,
	argvars(I1,N,X,Vs1,Vs2).
*/

	
vars(X,Vs,Vs1) :-
        nonvar(X),
        X =.. [_|Args],
        argvars(Args,Vs,Vs1).
 
argvars([],Q,Q).
argvars([X|Xs],Vs,Vs2) :-
        vars(X,Vs,Vs1),
        argvars(Xs,Vs1,Vs2).
 
insertvar(X,[],[X]).
insertvar(X,[Y|Vs],[Y|Vs]) :-
        X == Y,
        !.
insertvar(X,[Y|Vs],[Y|Vs1]) :-
        insertvar(X,Vs,Vs1).

prolog_meta(X^B,X^B1,B,B1).
prolog_meta(setof(X,P,S),setof(X,P1,S),P,P1).
prolog_meta(bagof(X,P,S),bagof(X,P1,S),P,P1).
prolog_meta(bagof1(X,P,S),bagof1(X,P1,S),P,P1).
prolog_meta(not(B),not(B1),B,B1).
prolog_meta(\+(B),\+(B1),B,B1).
prolog_meta(neg(B),neg(B1),B,B1).  % for BIM, to handle \+
prolog_meta(call(B),call(B1),B,B1).
prolog_meta(time(B),time(B1),B,B1).
prolog_meta((C; D),(C1; D1),(C,D),(C1,D1)).
prolog_meta((B -> C),(B1 -> C1),(B,C),(B1,C1)).

:- dynamic neg/1.

neg(P) :-
	\+ P.

remove(X,[X|Y],Y).
remove(X,[Y|Ys],[Y|Zs]) :-
        remove(X,Ys,Zs).


writeqBody((G,Gs)) :-
        !,
        write('      '),
        /*writeq(G), */
        write(G),
        write(','),nl,
        writeqBody(Gs).
writeqBody(G) :-
        write('      '),
        /*writeq(G), */
        write(G),
        write('.'),nl.


% Henk's version to handle quoted atoms

displayqClauses([]).
displayqClauses([Clause|Cls]) :-
	sp_writeClause(Clause),
        displayqClauses(Cls).



displayqRClauses([]).
displayqRClauses([Clause|Cls]) :-
	melt(Clause,C),
	sp_writeRclause(C),
        displayqRClauses(Cls).

writeqRBody((G,Gs)) :-
        !,
        /*writeq(G), */
        write(G),
        write(','),
        writeqRBody(Gs).
writeqRBody(G) :-
        /*writeq(G), */
        write(G),
        write('.'),nl.

variable(X) :-
	variable(X,_).

variable(X,N) :-
        atom(X),
        name('X',[A]),
        name(X,[A|Num]),
	name(N,Num),
	integer(N).

depth_k(0,X,X) :-
	var(X),
	!.
depth_k(0,C,C) :-
	C =.. [C],
	!.
depth_k(0,_,_).
depth_k(K,T1,T2) :-
	K > 0,
	depth_k_abstract(K,T1,T2).

depth_k_abstract(_,V,V) :-
	var(V),
	!.
depth_k_abstract(K,T1,T2) :-
	T1 =.. [P|Xs],
	K1 is K-1,
	depth_k_all(K1,Xs,Ys),
	T2 =.. [P|Ys].

depth_k_all(_,[],[]).
depth_k_all(K,[X|Xs],[Y|Ys]) :-
	depth_k(K,X,Y),
	depth_k_all(K,Xs,Ys).


/*
instance_of(X,Y) :-
	melt(X,Mx),
	melt(Y,Mx),
	canonical(Mx),
	X == Mx.
*/
	
/* instanceof(X,Y):  X in ground representation, Y a ground instance of X */

instance_of(X,Y) :-
        inst(X,Y,[],_).

inst(V,X,[],[(V/X)]) :-
        variable(V).
inst(V,X,[(V/X)|S],[(V/X)|S]) :-
        variable(V).
inst(V,X,[(M/Y)|S],[(M/Y)|S1]) :-
        variable(V),
        V \== M,
        inst(V,X,S,S1).
inst(T,T1,S,S1) :-
        T =.. [F|Xs],
        \+ variable(T),
        functor(T,F,N),
        functor(T1,F,N),
        T1 =.. [F|Ys],
        inst_args(Xs,Ys,S,S1).

inst_args([],[],S,S).
inst_args([X|Xs],[Y|Ys],S,S2) :-
        inst(X,Y,S,S1),
        inst_args(Xs,Ys,S1,S2).



replaceterm(V,X,T,X) :-
        T == V,
        !.
replaceterm(_,_,T,T) :-
        var(T),
        !.
replaceterm(V,X,T,T1) :-
        nonvar(T),
        T =.. [F|Us],
        replaceall(V,X,Us,Vs),
        T1 =.. [F|Vs].

replaceall(V,X,[U|Us],[W|Ws]) :-
        replaceterm(V,X,U,W),
        replaceall(V,X,Us,Ws).
replaceall(_,_,[],[]).


