/* determinate unfolding with user-controlled unfolding choices */

resultant(catom(G,(U,C)),P,(R,U2,C),K0) :-
	ref_clause(G,B),
	detsolve(U,C,U1),
	derivation(B,R,U1,C,K0),
	unfold_body(U1,C,U2),
	non_empty_rul(U2,C,K0),
	no_Deadend(R),
	return_abstraction(G,P).	% return the trace term 
					% or other abstraction

rul_consistent(U,C,K) :-
	rul_consistent(U,C,_,K).

rul_consistent(U,C,U2,K) :-
	unfold_body(U,C,U1),
	non_empty_rul(U1,C,K),
	detsolve(U1,C,U2).
	
detsolve(true,_,true) :-
	!.
detsolve((B,Bs),C,Ds) :-
	det_ruldef(B,C,P),
	melt(P, [(B :- B2)]),
	!,
	joingoals(B2,Bs,Bs1),
	detsolve(Bs1,C,Ds).
detsolve((B,Bs),C,(B,Ds)) :-
	!,
	detsolve(Bs,C,Ds).
detsolve(B,C,Ds) :-
	det_ruldef(B,C,P),
	melt(P, [(B :- B2)]),
	!,
	detsolve(B2,C,Ds).
detsolve(B,_,B).

det_ruldef(any(_),_,[(any('X0') :- true)]) :-
	!.
det_ruldef(B,C,Clauses) :-
	B =.. [T|_],
	def(T/1,Clauses,C).

%return_abstraction(statement(_,[_|_],[p(L,_)|_],_),L) :-
%	!.
%return_abstraction(statement(_,[V|_],[p(L,_)|_],_),(L,V)) :-
%	!.
return_abstraction(statement(_,[_,V|_],[p(L,_)|_],_),(L,V)) :-
	!.
return_abstraction(byte(_,_,_,_,[p(L,_)|_],_),L) :-
	!.
return_abstraction(path(_,A,_,_,_),A) :-
	!.
return_abstraction(wpath(_,A,_,_,_),A) :-
	!.
%return_abstraction(code_id_cont(_,L,_,[V|_],_),(L,V)) :-
%	!.
return_abstraction(G,P) :-
	arg(1,G,P).			% Trace term is 1st argument


non_empty_rul(U1,C,K0) :-
	intersect_body(U1,C,_,_,K0,_).

unary_constraints([],true,true).
unary_constraints([A],B,U) :-
	functor(B,T,1),
	U =.. [T,A].
unary_constraints([A1,A2|As],(B,Bs),(U,Us)) :-
	functor(B,T,1),
	U =.. [T,A1],
	unary_constraints([A2|As],Bs,Us).
	
/* Determinate unfolding only */

derivation(true,true,_,_,_) :-
	!.
derivation(Gs,Gs1,U,C,K) :-
	derivation_step(Gs,Gs2,true,Suspend,U,C,K),
	no_Deadend(Gs2),
	rul_consistent(U,C,U1,K),
	derivation_continue(Suspend,Gs2,Gs1,U1,C,K).

derivation_continue(yes,Gs,Gs,_,_,_).
derivation_continue(no,Gs2,Gs1,U,C,K) :-
	derivation(Gs2,Gs1,U,C,K).
	
derivation_step((\+G,Gs),Gs,_,no,_,_,_) :-
	unfoldable(\+G),
	!,
	\+ sp_solve(G).
derivation_step((B,Gs),Gs,_,no,_,_,_) :-
	sp_builtin(B),
	unfoldable(B),
	!,
	call(B).
derivation_step((B,Gs),(B,Gs1),Hs,Susp,U,C,K) :-
	sp_builtin(B),
	!,
	derivation_step(Gs,Gs1,(B,Hs),Susp,U,C,K).
derivation_step((B,Gs),(B,Gs1),Hs,Susp,U,C,K) :-
	suspend_goal((B,Gs),U,C,K),
	!,
	derivation_step(Gs,Gs1,(B,Hs),Susp,U,C,K).
derivation_step((B,Gs),Gs1,_,no,_,_,_) :-
	!,
	ref_clause(B,R),
	joingoals(R,Gs,Gs1).
derivation_step(\+B,true,_,no,_,_,_) :-
	unfoldable(\+B),
	!,
	\+ sp_solve(B).
derivation_step(B,true,_,no,_,_,_) :-
	sp_builtin(B),
	B \== true,
	unfoldable(B),
	!,
	call(B).
derivation_step(B,B,_,yes,_,_,_) :-
	sp_builtin(B),
	!.
derivation_step(B,B,Hs,yes,U,C,K) :-
	suspend_goal((B,Hs),U,C,K),
	!.
derivation_step(B,R,_,no,_,_,_) :-
	ref_clause(B,R).

% user defined suspend conditions, if any, go here

%suspend_goal(q(_,_),_,_,_) :-
%	!.
%suspend_goal(r1(_,_,_,_),_,_,_) :-
%	!.
%suspend_goal(as(_,_,_,_),_,_,_) :-
%	!.
%suspend_goal((q(_,_),_),_,_,_) :-
%	!.
%suspend_goal((r1(_,_,_,_),_),_,_,_) :-
%	!.
%suspend_goal((as(_,_,_,_),_),_,_,_) :-
%	!.
suspend_goal(statement(_,_,[p(_,_)|_],_),_,_,_) :-
	!.
suspend_goal((statement(_,_,[p(_,_)|_],_),_),_,_,_) :-
	!.
suspend_goal(byte(_,_,_,_,[p(_,_)|_],_),_,_,_) :-
	!.
suspend_goal((byte(_,_,_,_,[p(_,_)|_],_),_),_,_,_) :-
	!.
suspend_goal(wpath(_,_,_,_),_,_,_) :-
	!.
suspend_goal((wpath(_,_,_,_),_),_,_,_) :-
	!.
suspend_goal(path(_,_,_,_),_,_,_) :-
	!.
suspend_goal((path(_,_,_,_),_),_,_,_) :-
	!.
suspend_goal(Byte,_,_,_) :-
	functor(Byte,F,_),
	name(F, [98,121,116,101,95|_]),
	!.
suspend_goal((Byte,_),_,_,_) :-
	functor(Byte,F,_),
	name(F, [98,121,116,101,95|_]),
	!.
suspend_goal(B,U,C,K) :-
	nondeterminate(B,U,C,K).

nondeterminate((G,Gs),U,C,K) :-
	!,
	bagof(Bs, (G,Gs,B,U)^(
		ref_clause(G,B),
		joingoals(B,Gs,Bs),
		no_Deadend(Bs),
		rul_consistent(U,C,K)),[_,_|_]).
nondeterminate(G,U,C,K) :-
	bagof(B, (G,U)^(
		ref_clause(G,B),
		no_Deadend(B),
		rul_consistent(U,C,K)),[_,_|_]).




