% JPG - 20.10.93
% Compute ground-any modes for the success set of a program

%:- setdebug.
 
tp_fix(P) :-
	time(tp_fix1(P,F)),
	display_fix(F).

tp_fix1(P,F) :-
	strong_connected_components(P,G),% compute the dependency ordering
	tp_fix_groups(G,[],F).

tp_fix_groups([],F,F) :-
	!.
tp_fix_groups([(R,G)|Gs],F,F2) :-
	nonrec(R),
%	write(G),nl,
	!,
	init(G,R0),
	tp_alpha(G,[[]],F,R0,R2),  % no need to iterate if non-rec.
	app(F,R2,F1),
	tp_fix_groups(Gs,F1,F2).
tp_fix_groups([(_,G)|Gs],F,F2) :-
%	write(G),nl,
	init(G,R0),
	tp_iterate([[]],G,F,R0,R2),
	app(F,R2,F1),
	tp_fix_groups(Gs,F1,F2).

nonrec(non_recursive).

%nonrec(_) :-
%	!,fail.

%nonrec(X) :-
%	\+ recursive_pred(X).

recursive_pred(X) :-
	memb2(P,X),
	recursive(P),
	!.

% tp_fix(Ps,R) :-
% 	init(Ps,R0),
% 	tp_iterate([[]],Ps,R0,R,10,_).

init([_|Ps],[[]|R]) :-
	init(Ps,R).
init([],[]).

tp_iterate([],_,_,R,R) :-
	!.
tp_iterate(Nofix,Ps,Acc,R1,R) :-
	% writestate(R1),
	app(R1,Acc,D),
	tp_alpha(Ps,Nofix,D,R1,R2),
	% countdefs(R2),
	check_fix(Ps,R1,R2,Fix),
	!,
	tp_iterate(Fix,Ps,Acc,R2,R).

countdefs([R|Rs]) :-
	sp_length(R,N),write(N),write(' '),
	!,
	countdefs(Rs).
countdefs([]) :-
	nl.


	
	
tp_alpha([],_,_,[],[]).
tp_alpha([P|Ps],Nofix,R,[PR|PRs],[UP|R1]) :-
	user_clauses(P,Cls),
	infer_proc(Cls,Nofix,R,As),
	ub_proc(As,PR,UP),
	tp_alpha(Ps,Nofix,R,PRs,R1).

infer_proc([(_ :- B)|Cls],Nofix,R,As) :- 	% if no predicate in the body
	nochange(B,Nofix),			% changed, skip clause
	!,
	infer_proc(Cls,Nofix,R,As).
infer_proc([(H :- B)|Cls],Nofix,R,[S|As]) :-
	tsolve(B,R,T),
	unfold_body(T,T1),
	!,
	headtype(H,T1,S), 	
	infer_proc(Cls,Nofix,R,As).
infer_proc([_|Cls],Nofix,R,As) :-		% if no solution in R.
	infer_proc(Cls,Nofix,R,As).
infer_proc([],_,_,[]).

nochange(_,[[]]) :-  	% everything handled on 1st iteration
	!,
	fail.
nochange(true,_) :-
	!.
nochange(B,Nofix) :-
	preds(B,Ps),
	\+ commonpred(Ps,Nofix).

preds(true,[]) :-
	!.
preds((B,Bs), [P/N|Ps]) :-
	!,
	functor(B,P,N),
        preds(Bs,Ps).
preds(B,[P/N]) :-
	functor(B,P,N).

commonpred(X,Y) :-
	memb2(Z,X),
	memb2(Z,Y).
	
tsolve(true,_,true) :-
	!.
tsolve((B,Bs),R,T) :-
	!,
	tsolve(B,R,T1),
	tsolve(Bs,R,T2),
	joingoals(T1,T2,T).
tsolve(B,_,Bs) :-
        sp_builtin(B),
	builtin_tsolve(B,Bs),
        !.
tsolve(B,R,T1) :-
	functor(B,S,N),
	headclause(S/N,R,(H :- T)),
	melt((H :- T),(B :- T1)).

builtin_tsolve(X is Y, (ground(X),ground(Y))) :-
	!.
builtin_tsolve(X < Y, (ground(X),ground(Y))) :-
	!.
builtin_tsolve(X =< Y, (ground(X),ground(Y))) :-
	!.
builtin_tsolve(X > Y, (ground(X),ground(Y))) :-
	!.
builtin_tsolve(X >= Y, (ground(X),ground(Y))) :-
	!.
builtin_tsolve(B, Bs) :-
	any_term(B,Bs).

any_term(B,Bs) :-
        vars(B,Vs),
        any_vars(Vs,Bs).

any_vars([],true) :-
        !.
any_vars([X],any(X)) :-
        !.
any_vars([X|Xs],(any(X),Bs)) :-
        any_vars(Xs,Bs).



unfold_body(true,true) :-
	!.
unfold_body((B,Bs),C) :-
	!,
	unfold_body(B,C1),
	unfold_body(Bs,C2),
	joingoals(C1,C2,C).
unfold_body(ground(X),D) :-
	vars(X,Vs),
	all_ground(Vs,D).
unfold_body(any(X),D) :-
	vars(X,Vs),
	all_any(Vs,D).

all_ground([],true).
all_ground([X|Xs],(ground(X),Gs)) :-
	all_ground(Xs,Gs).
	
all_any([],true).
all_any([X|Xs],(any(X),Gs)) :-
	all_any(Xs,Gs).
	

headtype(T, T1, (X :- Bs)) :- 
	functor(T,F,N),
	functor(X,F,N),
	headargs(T,X,N,1,Bs,T1),
	canonical((X :- Bs)).
	
headargs(_,_,N,J,true,_) :-
	J > N,
	!.
headargs(T,X,N,J,(G,Gs),T1) :-
	arg(J,T,Tj),
	arg(J,X,Xj),
	vars(Tj,Vjs),
	groundarg(Vjs,Xj,T1,G),
	J1 is J+1,
	headargs(T,X,N,J1,Gs,T1).

groundarg(Xs,X,T,ground(X)) :-
	\+ nonground(Xs,T),
	!.
groundarg(_,X,_,any(X)).

nonground(Xs,T) :-
	memb2(Y,Xs),
	\+ bodyground(Y,T).

bodyground(Y,(ground(Z),_)) :-
	Y == Z,
	!.
bodyground(Y,(_,Gs)) :-
	!,
	bodyground(Y,Gs).
bodyground(Y,ground(Z)) :-
	Y == Z. 


check_fix([],[],[],[]).
check_fix([_|Ps],[RP|R],[RP1|S],Qs) :-
        fixpoint(RP,RP1),
	!,
        check_fix(Ps,R,S,Qs).
check_fix([P|Ps],[_|R],[_|S],[P|Qs]) :-
        check_fix(Ps,R,S,Qs).

fixpoint(X,X).
	
ub_proc([],[],[]).
ub_proc([(S :- B)|As],[],UP) :-
	definite_ground(S,[(S :- B)|As],UP).
ub_proc(As,[(S :- B)],UP) :-
	definite_ground(S,[(S :- B)|As],UP).


definite_ground(S,Cls,[(S :- B2)]) :-
	functor(S,_,N),
	ub_bodyground(S,Cls,B1,N,1),
	trimtrue(B1,B2).

trimtrue(true,true) :-
	!.
trimtrue((B,true),B) :-
	!.
trimtrue((B,Bs),(B,Bs1)) :-
	trimtrue(Bs,Bs1).

ub_bodyground(_,_,true,N,I) :-
	I > N.
ub_bodyground(S,Cls,(ground(Si),Bs),N,I) :-
	I =< N,
	arg(I,S,Si),
	def_ground(Si,Cls),
	!,
	I1 is I+1,
	ub_bodyground(S,Cls,Bs,N,I1).
ub_bodyground(S,Cls,(any(Si),Bs),N,I) :-
	I =< N,
	arg(I,S,Si),
	I1 is I+1,
	ub_bodyground(S,Cls,Bs,N,I1).

def_ground(Si,Cls) :-
	\+ poss_any(Si,Cls).

poss_any(Si,Cls) :-
	memb2((_ :- B),Cls),
	arg_is_any(Si,B).
	
arg_is_any(Si,(any(Si),_)) :-
	!.
arg_is_any(Si,(_,Bs)) :-
	!,
	arg_is_any(Si,Bs).
arg_is_any(Si,any(Si)).

headclause(P/N,R,(H :- B)) :-
	memb2([(H :- B)],R),
	functor(H,P,N).



:- dynamic traceon/0.

writestate(R) :-
	traceon,
	write(R),
	read(_),
	!,
	nl.
writestate(_) :-
	write('*'),
	sp_ttyflush.

display_fix(Cls) :-
        nl,write('Enter filename: '),
        read(F),
        reg_filewrite(F,Cls).

reg_filewrite(tty,Cls) :-
        !,
	nl,write('Definite Sharing Approximation:'),nl,nl,
        displayqTypes(Cls),
	nl,nl.
reg_filewrite(F,Cls) :-
        tell(F),
        displayqTypes(Cls),
        told.

/* write the program to current output file  */

displayqTypes([[]|Ps]) :-
	!,
	displayqTypes(Ps).
displayqTypes([P|Ps]) :-
	displayqRClauses(P),
	displayqTypes(Ps).
displayqTypes([]).



subatoms(T,Vs) :-
        subatoms(T,[],Vs).

subatoms(X,Vs,Vs1) :-
        atomic(X),
        !,
        insertatom(X,Vs,Vs1).
subatoms(X,Vs,Vs1) :-
        var(X),
        !,
        insertatom(X,Vs,Vs1).
subatoms(X,Vs,Vs1) :-
        nonvar(X),
        X =.. [_|Args],
        argsubatoms(Args,Vs,Vs1).
 
argsubatoms([],Q,Q).
argsubatoms([X|Xs],Vs,Vs2) :-
        subatoms(X,Vs,Vs1),
        argsubatoms(Xs,Vs1,Vs2).
 
insertatom(X,[],[X]).
insertatom(X,[Y|Vs],[Y|Vs]) :-
        X == Y,
        !.
insertatom(X,[Y|Vs],[Y|Vs1]) :-
        insertatom(X,Vs,Vs1).

constant(K) :-
	name(K,[A|_]),
	name(K1,[A]),
	K1 \== 'A'.
variable(K) :-
	name(K,[A|_]),
	name('A',[A]).
