%:- alldynamic.
%:- setdebug.
%:- dynamic user_proc/3.

strong_connected_components(Ps,PG) :-
	dependencies(Ps,Dlist),
	depend_sort(Dlist,Dlist1),
	partial_order(Dlist1,PG).
	%time(dependencies(Ps,Dlist)),
	%time(depend_sort(Dlist,Dlist1)),
	%time(partial_order(Dlist1,PG)).
	
% for each predicate compute set of dependent predicates

dependencies([],[]) :-
	!.
dependencies([P|Ps],[(N,R,PDs)|Dlist]) :-
	reachable(P,PDs,R),
	sp_length(PDs,N),
	write(N),write(' '),
	dependencies(Ps,Dlist).

depend_sort([],[]) :-
	!.	
depend_sort([X|Xs],Ys) :-
	depend_sort(Xs,Ys,[X]).
	
% insertion sort

depend_sort([X|Xs],Ys,Ws) :-
	insert_asc(X,Ws,Vs),
	depend_sort(Xs,Ys,Vs).
depend_sort([],L,L).

insert_asc(X,[X|Xs],[X|Xs]) :-
	!.
insert_asc(X,[Y|Ys],[X,Y|Ys]) :-
	X @< Y,
	!.
insert_asc(X,[Y|Ys],[Y|Ys1]) :-
	insert_asc(X,Ys,Ys1).
insert_asc(X,[],[X]).
	
partial_order(X,Y) :-
	partial_order(X,Y,[]).

partial_order([(_,R,X)|Xs],[(R,U)|Us],V) :-
	set_union(X,V,W),
	set_difference(X,V,U),
	U \== [],
	!,
	partial_order(Xs,Us,W).
partial_order([_|Xs],Us,V) :-
	partial_order(Xs,Us,V).
partial_order([],[],_).
	
set_difference(X,[],X).
set_difference([],_,[]).
set_difference([X|Xs],Ys,Ws) :-
	memb1(X,Ys),
	!,
	set_difference(Xs,Ys,Ws).
set_difference([X|Xs],Ys,[X|Ws]) :-
	!,
	set_difference(Xs,Ys,Ws).
	
set_union([],Y,Y) :-
	!.
set_union([X|Xs],Ys,Zs) :-
	insert_asc(X,Ys,Ys1),
	!,
	set_union(Xs,Ys1,Zs).

subset(X,Y) :-
	\+ excess_member(X,Y).

excess_member(X,Y) :-
	memb1(Z,X),
	\+ memb1(Z,Y).

	
/* get all predicates reachable from a predicate */	

reachable(T,Ps,R) :-
	iterate_preds([T],Ps1),
	close_list(Ps1),
	classify_rec(T,Ps1,Ps,R).

classify_rec(T,Ps,Ps,recursive) :-
	memb1(T,Ps),
	!.
classify_rec(T,Ps,[T|Ps],non_recursive).


iterate_preds([],_) :-
	!.
iterate_preds(Ts,Ss) :-
	each_pred(Ts,Ss,More),
	iterate_preds(More,Ss).

each_pred([],_,_) :-
	!.
each_pred([T/N|Ts],Ss,M) :-
	user_clauses(T/N,Cls),
	body_preds(Cls,Ss,M),
	each_pred(Ts,Ss,M).

body_preds([(_ :- B)|Cs],Ss,M) :-
	bodylits(B,Ss,M),
	body_preds(Cs,Ss,M).
body_preds([],_,_).

bodylits((B,Bs),Ss,M) :-
	\+ sp_builtin(B),
	!,
	functor(B,T,N),
	insertT(T/N,Ss,M),
	bodylits(Bs,Ss,M).
bodylits((_,Bs),Ss,M) :-
	!,
	bodylits(Bs,Ss,M).
bodylits(true,_,_) :-
	!.
bodylits(B,Ss,M) :-
	\+ sp_builtin(B),
	!,
	functor(B,T,N),
	insertT(T/N,Ss,M).
bodylits(_,_,_).

