%*****************************************************************
%*			Logic Programing.			*
%*			 Course Project				*
%*	by							*
%*		Diber Alex, I.D. 30673577 			*
%*								*
%*								*
%*****************************************************************
%
%DESCRIPTION:
%
%This project is written on QProlog, and perform some analysis and
%transformation of prolog program. It work as follow :
%
%	1. read prolog program  from file into internal structure.
%	2. build dependency tree of this program
%	3. determine strong connected components of this graph.
%	4. determine recursive predicates of the program
%	5. perform some transformation of initial program and write
%		the transformed program to file.
%
%
%FILES:
%
%The project consist of the following files :
%	read   	- predicates that deal with reading and parsing prolog
%			program and building dependency tree.
%	write  	- predicates that deal with transforming the program and
%			writing its' to file
%	scc	- algorithm for detection strong connected components
%	run	- some simple functions, detection recursive predicates,
%		  and the top level function
%			run(InputFileName, OutputFileName)
%
%ALGORITHMS:
%
%	Algorithm for detection strong connected components is
%taken from  R. Tarjan: SIAM J. Computing; Vol 1. No 2. p. 146-160.
%
%DATA STRUCTURES:
%
%The next data structures used by this program :
%
%	1. When input file with prolog program is read (by
%read_file(File, ClauseList) function) , the program is stored
%format of list of clauses , where each clause of program like
%	p:-q,r,s
%included in list in form
%	(p,[q,r,s]).
%
%	2. dependency graph represented as list of pairs
%	(NodeName,[List of names of adjacent nodes])
%for example , if program is
%	p:-q,r,p.
%	q:-r.
%	r:-q.
%the corresponding dependency  graph is:
%	[(p,[q,r,p]),(q,[r]),(r,[q])]
%
%
%	3.prep(Graph,NewGraph) that change format of graph
%representation to form that more appropriative for DFS.
%It work as follow : if input graph was
%		[	(p,[q,r])	(q,[r])		(r,[])	]
%first , it build the next structure :
%		[ 	(p,_,_,[_,_])	(q,_,_,[_])	(r,_,_,[])]
%where underline (_) is not bound variable.
%During DFS, the first two _ of each touple will be bound to
%NUMBER and LOWLINK corresponding to this node. The 4th element
%of touple is list , with same number of elements ,that list
%in old touple had (i.e. _ for each edge).
%Second, each _ in the list  is bind to the touple,
%corresponding to node the arc point to it:
%
%		[ 	(p,_,_,[_,_])	(q,_,_,[_])	(r,_,_,[])]
%               			|       ^	|      ^^
%				| |	|	+------+|
%				+-------+		|
%				  |			|
%				  +---------------------+
%
%	(arcs symbol binding of logic variable to value)
%
%
%OUTPUT FILE FORMAT :
%	
%In output file, all term
%	q
%of input program is changed by 	
%	scci(q)
%where scci is name for strong connected component to that
%q belong.
%The  clause
%	order([scc1,scc2,...,sccn]).
%is added to output. It insert full order on set of connected component.
%
%
%ALGORITHM COMPLEXITY :
%
%Suppose that input program consist of C clauses, T terms, and
%dependency graph consist of V nodes , E arcs, S strong connected components,
%and maximum degree of node in the graph is D, maximum size of SCC is M.
%
%1. When dependency graph are built (see ins_edges, make_graph, insert_edge,
%insert_vertex, insert_new  functions), each new node (T times) are inserted into
%list of nodes (max length V), so complexity is O(VT). Analogous, when arcs inserted
%the   complexity is O(VE+ED).
%
%2. function prep(Graph,NewGraph) work O(V+VE) time.
%
%3. SCC algorithm work O(V+VE) time because:
%	forest_dfs, dfs, pop_component, pop_if_need - called once for each node (V)
%	for_all - called once for each arc (E).
%		for_all call member(V,Stack) that can work up to O(V)
%, so result time is O(V+VE). This can be improved by changing member(,) call
%by a flag.
%
%4. writing of output file take O(TS+T+S+C) = O(TS)
%5. recursivity checking O(VM)	
%
%EXAMPLE :
%
%	ivory.bgu.ac.il 142 > cat >test
%	r:-q,p,r.
%	q:-p.
%	p:-q.
%
%	ivory.bgu.ac.il 143 > qprolog
%	Quintus Prolog Release 3.0 (DECstation, Ultrix 4.0)
%	Copyright (C) 1990, Quintus Computer Systems, Inc.  All rights reserved.
%	2100 Geng Road, Palo Alto, California U.S.A. (415) 813-3800	
%
%	| ?- [read].
%	% compiling file /users/black3/master/diber/logpro/proj/read
%	% read compiled in module user, 0.283 sec 1,720 bytes
%
%	yes
%	| ?- [write].
%	% compiling file /users/black3/master/diber/logpro/proj/write
%	% write compiled in module user, 0.250 sec 1,712 bytes
%
%	yes
%	| ?- [scc].
%	% compiling file /users/black3/master/diber/logpro/proj/scc
%	* Singleton variables, clause 1 of dfs/7: Name
%	* Singleton variables, clause 2 of prep/3: Back, No
%	% scc compiled in module user, 0.467 sec 2,344 bytes
%
%	yes
%	| ?- [run].
%	% compiling file /users/black3/master/diber/logpro/proj/run
%	% run compiled in module user, 0.117 sec 608 bytes
%
%	yes
%	| ?- run(test,result).
%
%	yes
%	| ?-
%
%	ivory.bgu.ac.il 144 > cat result
%	/*
%	List of SCCs:
%	[[r/0],[p/0,q/0]]
%	List of recursive predicates:
%	[r/0]
%	*/
%	order([scc1,scc2]).
%	scc1(r) :-
%	        scc2(q) ,
%	        scc2(p) ,
%	        scc1(r)
%	.
%	scc2(q) :-
%	        scc2(p)
%	.
%	scc2(p) :-
%	        scc2(q)
%	.
%
%
%
%
%
%
%
/****************************************************************
*			Logic Programing.			*
*			 Course Project				*
*	by							*
*		Diber Alex, I.D. 30673577 			*
*								*
*								*
*****************************************************************/

/*
	This file contain functions that read prolog program from a
	file and build from it dependency tree.
	
	Top level functions are :
		read_file(File, ClauseList)
	where
		File - file name to read.
		ClauseList - list of clauses of program , where each clause
	of program like p:-q,r,s included in list in form (p,[q,r,s]).
		
		make_graph(ClauseList,Graph)
	where
		ClauseList - like above.
		Graph - dependency graph represented as list of
	pairs (NodeName,[List of names of adjesent nodes]) .


 If program is:
	p:-q,r.
	p:-r.
	q:-r.
	r:-q.
 then ClauseList is
	[(p,[q,r]),(p,[r]),(q,[r]),(r,[q])]
 and Graph is
	[(p,[q,r])(q,[r])(r,[q])]			
		
*/

read_file(File, ClauseList) :-
	see(File),
	read(X),
	read_clauses(X,ClauseList).

read_clauses(X,[]):-
	X = end_of_file,!,
	seen.
read_clauses(X,[XX|Xs]):-
	parse(X,XX),
	read(Y),!,
	read_clauses(Y,Xs).

/* parse clause ( split it as left-hand-side:-right-hand-side ) */
parse(X,(A,L)) :-
	(X) = (A:-B) , !, /* clause */
	parse_rhs(B,L).
	
parse(X,(X,[])). /* fact */

/* parse right-hand-side into list of terms */
parse_rhs(X,[A|L1]) :-
	(X) = (A,B) , ! ,
	parse_rhs(B,L1).
parse_rhs(X,[X]).




/*
make_graph(ClauseList,Graph)
*/

make_graph([],[]).
make_graph([(V,List)|R],Graph):-
	make_graph(R,G1),
	make_name(V,VV),
	insert_vertex(VV,G1,G2 ),
	ins_edges(VV , List , G2 , Graph ).
	
/*	
ins_edges(VV , EdgeList , Graph , NewGraph )
*/
ins_edges(_ , [] , G , G ).
ins_edges(VV , [A|Rest] , G , NewG ):-
	make_name(A,AA),
	insert_vertex(AA,G,G1),
	insert_edge(VV,AA,G1,G2),!,
	ins_edges(VV , Rest , G2 , NewG ).
	

/*
make_name(Term,Name)
make node name Name that correspond to term Term of program
*/
make_name(V,Name):-
	functor(V, TermName , Arity),
	name(TermName,NN),
	name(Arity,AA),
	append(NN,"/",NNN),
	append(NNN,AA,NNAA),
	name(Name,NNAA).


/*	
insert_edge(V1,V2,Graph,NewGraph)
insert new arc (V1,V2) into graph Graph.
*/

insert_edge(V1,V2,[(V1,L)|R], [(V1,LL)|R]):- !,
	insert_new(V2,L,LL).
insert_edge(V1,V2,[V|R], [V|R1]):- !,
	insert_edge(V1,V2,R, R1).

/*
insert_vertex(V,Graph,NewGraph)
insert new node into graph Graph.
*/	

insert_vertex(V,G,G):-
	G = [(V,_)|_],!.
insert_vertex(V,[],[(V,[])]):-!.
insert_vertex(V,[W|Vs],[W|NewVs]):-insert_vertex(V,Vs,NewVs).

/*
insert_new(Element,List,NewList).
insert new element into list .
*/

insert_new(Element,[Element|Rest],[Element|Rest]):-!.
insert_new(Element,[],[Element]):-!.
insert_new(Element,[F|Rest],[F|NewRest]):-
	insert_new(Element,Rest,NewRest).


/****************************************************************
*                       Logic Programing.                       *
*                        Course Project                         *
*       by                                                      *
*               Diber Alex, I.D. 30673577                       *
*                                                               *
*                                                               *
*****************************************************************/

/*
	This file contain functions that write results of
	computing SCC into file.
	
	Top level function is
		write_results(File , ClauseList, Graph, Res)
	where
		File - file name to write
		ClauseList - initial prolog program (in form of clauses' list ),
 	that will be modified.
		Graph - Dependency graph.
		Res - List of SCCs.

*/

write_results(File , ClauseList, G , Res):-
	open(File ,write,A),
	write(A,'/*'),  nl(A),

	write(A,'List of SCCs:'),nl(A),	
	write(A,Res) ,  nl(A) ,

	recursive(G,L),
	write(A,'List of recursive predicates:'),nl(A),
	write(A,L),nl(A),

	write(A,'*/'),  nl(A),

	write_order(A,Res),
	write_clauses(A,ClauseList,Res),
	close(A).


/*
write_clauses(A,ClauseList,Res)
	write modified list of clauses ClauseList into file A
*/
write_clauses(_,[],_).
write_clauses(A,[Clause|List],Res):-
	write_clause(A,Clause,Res),!,
	write_clauses(A,List,Res).

/*
write_clauses(A,Clause,Res)
	write modified clause Clause into file A
*/
write_clause(A,(Head,Tail),Res) :-
	make_name(Head,HeadName),
	find_scc(HeadName,Res,I),
	write(A,scc), write(A,I) , write(A,'(') ,
	write(A,Head), write(A,') :-	') ,nl(A),
	write_tail(A,Tail,Res),
	write(A,'.'),nl(A).
/*
	write tail of clause (that after :- )
*/
write_tail(_,[],_).
write_tail(A,[Term|List],Res):-
	make_name(Term,TermName),
	find_scc(TermName,Res,I),
	write(A,'	') ,write(A,scc), write(A,I) , write(A,'(') ,
	write(A,Term),
	write(A,')	'),
	write_psik(A,List),
	nl(A),
	write_tail(A,List,Res).

/*
	write ',' if not end of list
*/
write_psik(_,[]) :- ! .
write_psik(A,_) :- write(A,',') .


	
/*
	write "order([scc1,scc2])."
*/	
write_order(A,Res):-
	write(A,'order(['),
	write_o(A,Res,1),
	write(A,']).'), nl(A).

write_o(_,[],_).
write_o(A,[_|R],I):-
	write(A,scc), write(A,I),
	write_psik(A,R),
	I1 is I+1,!,
	write_o(A,R,I1).


/*
find_scc(NodeName,SCCList,No)
	Find number of SCC in list SCCList , that contain
	node with name NodeName.
*/
find_scc(HeadName,[A|_],1):-
	member(HeadName,A),!.

find_scc(HeadName,[_|B],I):-
	find_scc(HeadName,B,I1),
	I is I1 + 1.










/****************************************************************
*                       Logic Programing.                       *
*                        Course Project                         *
*       by                                                      *
*               Diber Alex, I.D. 30673577                       *
*                                                               *
*                                                               *
*****************************************************************/


/*
	This file contain program that determine the strongly
connected componets of graph.

The algorithm that used compute SCCs by using DFS. During DFS for
each node of tree computed two numbers : NUMBER - the number of
the nodes in order they are reached during the DFS , LOWLINK - the
smalest vertex which is in same component and reacheable by traversing
zero or more tree arcs followed at most one frond or cross-link.   	

The top level function is scc(Graph,Res) that receive graph Graph
( in format as discribed in file "read") and return list of SCCs Res.
It work as follow :
	1.)call to function prep(Graph,NewGraph) that change format of
graph representetion to form that more appropriative for DFS.
It work as follow :
	if input graph was
		[	(p,[q,r])	(q,[r])		(r,[])	]
	, first , it build the next structure :
		[ 	(p,_,_,[_,_])	(q,_,_,[_])	(r,_,_,[])]
	where underline (_) is not binded variable.
	During DFS, the first two _ of each touple will be bound to
	NUMBER and LOWLINK corresponding to this node. The 4th element
	of touple is list , with same number of elements ,that list
	in old touple had (i.e. _ for each edge).
	Second, each _ in the list  is binded to the touple,
	corresponding to node the arc point to it:
		[ 	(p,_,_,[_,_])	(q,_,_,[_])	(r,_,_,[])]
               			|       ^	|      ^^
				| |	|	+------+|
				+-------+		|
				  |			|
				  +---------------------+


	2.)call to the forest_dfs(...) function that run dfs algoritm for
		each yet not numbered node in graph.

*/


/*
scc(Graph,Res)
	Input : graph Graph
	Output: list of SCCs Res.
First of all it  change format of graph representation ( by call
to prep ) and then call the forest_dfs function )

*/
scc(Graph,Res):-
	prep(Graph,NewGraph),
	forest_dfs( NewGraph , [] , _ , 0 , _ , [] , Res).


/*
forest_dfs( V , Stack , NewStack , No , NewNo , Res , NewRes)
	run dfs algoritm for all not visited vertex.
*/

forest_dfs( [] , Stack , Stack , No , No , Res , Res).
forest_dfs( [V|R] , Stack , NewStack , No , NewNo , Res , NewRes):-
	V = (_,_,Num,_) ,
	var(Num),!,
	dfs( V , Stack , Stack1 , No , No1 , Res , Res1),
	forest_dfs( R , Stack1 , NewStack , No1 , NewNo , Res1 , NewRes).
forest_dfs( [_|R] , Stack , NewStack , No ,NewNo , Res , NewRes):-
	forest_dfs( R , Stack , NewStack , No, NewNo , Res , NewRes).

/*
dfs( V , Stack , NewStack , No , NewNo , Res , NewRes) :-
	run dfs algoritm from vertex V
*/

dfs( V , Stack , NewStack , No , NewNo , Res , NewRes) :-
	V = (Name,Edges,Num,Back) ,
	Num is No ,		/* mark this node as visited */
	push(Stack, V , S1),
	/* write(Name),nl, */
	NoNo is No+1,
	/* run dfs recursivly for all not marked childrens */
	for_all( Edges, S1, S2, NoNo, NewNo, Res, R1, No, Back),
	pop_if_need(Num,Back,S2,NewStack,R1,NewRes).	
		

/*
Run dfs for all not marked  nodes in List
for_all ( List , Stack , NewStack , No , NewNo , Res , NewRes , LowLink , NewLowLink)
*/

for_all( [] , Stack , Stack , No , No , Res , Res, LowLink , LowLink).

/* tree arc : run dfs and compute LowLink */
for_all( [V|Vs] , Stack , NewStack , No , NewNo , Res , NewRes, LowLink , NewLowLink):-
	V = (_,_,Num,Back),
	var(Num) , ! ,
	dfs( V , Stack , Stack1 , No , No1 , Res , Res1),
	min(Back,LowLink,LowLink1),!,
	for_all( Vs , Stack1 , NewStack , No1 , NewNo , Res1 , NewRes,LowLink1,NewLowLink).

/* frond or cross link : compute LowLink  only */
for_all( [V|Vs] , Stack , NewStack , No , NewNo , Res , NewRes, LowLink , NewLowLink):-
	V = (Name,_,Num,_),
	Num < No-1 ,
	member((Name,_,_,_),Stack),!, /* my be changed by addition a flag, for more effectivity   */
	min(Num,LowLink,LowLink1),!,
	for_all( Vs , Stack , NewStack , No , NewNo , Res , NewRes,LowLink1,NewLowLink).
/* rest of links : do nothing */
for_all( [_|Vs] , Stack , NewStack , No , NewNo , Res , NewRes, LowLink , NewLowLink):-
	for_all( Vs , Stack , NewStack , No , NewNo , Res , NewRes,LowLink,NewLowLink).


/*
pop_if_need(Num,Back,Stack,NewStack,Res,NewRes).	
	If current node is root of scc then pop it from stack
*/

pop_if_need(Num,Back,Stack,NewStack,Res,NewRes):-
	Num == Back , ! ,
	pop_component(Stack,Back,NewStack,[],R),
	/* write(pop),write(R),nl, */
	NewRes = [R|Res].
pop_if_need(_,_,S,S,R,R). /* thanks,  not need */



/*
pop_component(Stack,LowLink,NewStack, Component ,NewComponent)
	pop connected component NewComponent from stack Stack.
	
*/
pop_component([],_,[],C,C).
pop_component([V|S],Back,NewStack,C,NewC):-
	V = (Name,_,Num,_),
	Num >= Back,
	C1 = [Name|C],
	pop_component(S,Back,NewStack,C1,NewC).

pop_component(Stack,Back,Stack,C,C):-
	Stack = [(_,_,Num,_)|_],
	Num < Back.

/*****************************************************/

prep(List,NewList) :- prep(NewList,List,NewList).


/*

prep(Graph,List,NewList)

*/

prep(_,[],[]).
prep(G,[(Name,List)|Vs],[W|Ws]):-
	W=(Name,NewList,No,Back),
	prep(G,Vs,Ws),
	bind_arcs(G,List,NewList).

bind_arcs(_,[],[]).
bind_arcs(G,[V|Vs],[W|Ws]):-
	fnd(G,V,W),
	bind_arcs(G,Vs,Ws).

fnd([W|_],Name,W) :- W = (Name,_,_,_),!.
fnd([_|Ws],Name,W) :- fnd(Ws,Name,W).



/****************************************************************
*                       Logic Programing.                       *
*                        Course Project                         *
*       by                                                      *
*               Diber Alex, I.D. 30673577                       *
*                                                               *
*                                                               *
*****************************************************************/

/*
	run(In,Out)	
	Run the program.
	input :
		In - input file name
		Out = Output file name
*/
%run(In,Out) :- 	
%	time(do(In,ClauseList,Graph, Res)),
%	write_results(Out,ClauseList,Graph,Res).
%
%do(In,ClauseList,Graph, Res) :-
%	read_file(In,ClauseList),
%	make_graph(ClauseList,Graph),
%	scc(Graph, Res).

r :- 	run(test,res).		

run(In,Out) :- 	
	read_file(In,ClauseList),
	time(do(ClauseList,Graph,Res)),
	write_results(Out,ClauseList,Graph,Res).

do(ClauseList,Graph,Res) :-
	make_graph(ClauseList,Graph),
	scc(Graph, Res).

run1(In,Out) :- 	
	time(read_file(In,ClauseList)),
	time(make_graph(ClauseList,Graph)),
	time(scc(Graph,Res)),
	time(write_results(Out,ClauseList,Graph,Res)).

/*

recursive( Graph , RecNodes)
        input : Graph  ( in format as discribed in file "read")
        output : list of recursive nodes

*/
recursive([] , []).
recursive( [(Node,Edges)|Rest] , [Node|Nodes]) :-
        member(Node,Edges), ! ,
        recursive(Rest,Nodes).
recursive( [_|Rest] , Nodes) :-
        recursive(Rest,Nodes).




/*
	some very simple functions
*/

min(A,B,A) :- A < B, !.
min(_,B,B).

push(Stack,E,[E|Stack]).

member(A,[A|_]).
member(A,[_|B]):-member(A,B).

append([],X,X).
append([U|X],Y,[U|Z]) :-
	append(X,Y,Z).
