% JPG - 14.2.92
% Compute a regular approximation of a given program.
% For a description of the algorithm,
% see Gallagher and de Waal:  "Regular Approximations of
% Logic Programs and Their Uses";  CSTR-92-06.

%:- setdebug.
 
%:- alldynamic.

tp_fix(P) :-
	time(tp_fix1(P,F)),
	display_fix(F).

tp_fix1(P,F) :-
	strong_connected_components(P,G),% compute the dependency ordering
	nl,
	tp_fix_groups(G,[],F,10,K),
	show_useless_clauses(P,F,K,_),
	precision(P,F).

tp_fix_groups([],F,F,K,K) :-
	!.
tp_fix_groups([(R,G)|Gs],F,F2,K,K2) :-
	nonrec(R),
	write(G),nl,
	!,
	init(G,R0),
	tp_alpha(G,[[]],F,R0,R2,_,K,K1),  % no need to iterate if non-rec.
	app(F,R2,F1),
	tp_fix_groups(Gs,F1,F2,K1,K2).
tp_fix_groups([(_,G)|Gs],F,F2,K,K2) :-
	write(G),nl,
	init(G,R0),
	tp_iterate([[]],G,F,R0,R2,K,K1),
	app(F,R2,F1),
	tp_fix_groups(Gs,F1,F2,K1,K2).

nonrec(non_recursive).

recursive_pred(X) :-
	memb2(P,X),
	recursive(P),
	!.

init([_|Ps],[[]|R]) :-
	init(Ps,R).
init([],[]).

tp_iterate([],_,_,R,R,K,K) :-
	!.
tp_iterate(Nofix,Ps,Acc,R1,R,K,K3) :-
	writestate(R1),
	app(R1,Acc,D),
	tp_alpha(Ps,Nofix,D,R1,R2,Chs,K,K1),
	fold(R2,R3,Chs,K1,K2),
	countdefs(R3),
	%check_fix(Ps,R1,R3,Fix),
	check_fix(Ps,Chs,Fix),
	!,
	tp_iterate(Fix,Ps,Acc,R3,R,K2,K3).

countdefs([R|Rs]) :-
	sp_length(R,N),write(N),write(' '),
	!,
	countdefs(Rs).
countdefs([]) :-
	nl.

fold([],[],[],K,K).
fold([RP|R1],[RP|R3],[nochange|Chs],K,K2) :-
	!,
	fold(R1,R3,Chs,K,K2).
fold([RP|R1],[RP2|R3],[change|Chs],K,K2) :-
        toptype(RP,T),
        rdefs(T,RP1,RP),
        foldp(RP1,RP2,K,K1),
        fold(R1,R3,Chs,K1,K2).

	
%:- dynamic level/1.

level(1).

foldp(RP,RP3,K,K2) :-
	head_functors(RP,Fs),
        same_head_functors(T1,T2,Fs),
        %level(Level),
        depends(T1,T2,TPs,RP),
        %same_functor_structure(Level,T2,T1,TPs),
        !,
	recursive(T1,T2,TPs,RP,RP1,K,K1),
	toptype(RP1,T),
	rdefs(T,RP2,RP1),
	foldp(RP2,RP3,K1,K2).
foldp(R,R,K,K).

/*
head_functors([proc(_,_)|RP],Fs) :-
        head_funcs1(RP,Fs,[]).

head_funcs1([],Fs,Fs).
head_funcs1([proc(T/1,Cls)|Ps],Fs,Fs2) :-
        clause_head_functors(T,Cls,Fs,Fs1),
        head_funcs1(Ps,Fs1,Fs2).

clause_head_functors(_,[],Fs,Fs).
clause_head_functors(T,[(H :- _)|Cls],[head_func(T,F/N)|Fs],Fs1) :-
        H =.. [_,X],
        functor(X,F,N),
        clause_head_functors(T,Cls,Fs,Fs1).

same_head_functors(T1,T2,Zs) :-
        two_membs(head_func(T1,F),head_func(T2,F),Zs),
        T1 \== T2,
        \+ more_funcs(T1,T2,Zs),
        \+ more_funcs(T2,T1,Zs).

more_funcs(T1,T2,Zs) :-
        memb2(head_func(T1,F),Zs),
        \+ memb2(head_func(T2,F),Zs).

two_membs(X,Y,[X|Xs]) :-
        memb2(Y,Xs).
two_membs(X,Y,[_|Xs]) :-
        two_membs(X,Y,Xs).
*/

head_functors([],[]).
head_functors([proc(T/1,Cls)|Ps],[head_func(T,TFs)|Fs]) :-
        clause_head_functors(Cls,TFs),
        TFs = [_,_|_],		% at least two functors
        !,
        head_functors(Ps,Fs).
head_functors([_|Ps],Fs) :-
        head_functors(Ps,Fs).
        
clause_head_functors([],[]).
clause_head_functors([(H :- _)|Cls],[F/N|Fs]) :-
        H =.. [_,X],
        functor(X,F,N),
        clause_head_functors(Cls,Fs).

same_head_functors(T1,T2,Zs) :-
        two_membs(head_func(T1,Fs1),head_func(T2,Fs2),Zs),
        T1 \== T2,
        \+ more_funcs(Fs1,Fs2),
        \+ more_funcs(Fs2,Fs1).
        
more_funcs(Fs1,Fs2) :-
        memb2(X,Fs1),
        \+ memb2(X,Fs2).

two_membs(X,Y,[X|Xs]) :-
        memb2(Y,Xs).
two_membs(X,Y,[_|Xs]) :-
        two_membs(X,Y,Xs).



%recursive(T1,T2,TPs,RP,RP1,K,K) :-
%        subtype(T2,T1,TPs),
%        !,
%        rdefs(T2,T2Ps,TPs),
%        normalise_defs(TPs,T2,T1,TPs1),
%        replace_procs(T2Ps,TPs1,TPs2),
%        replace_procs(TPs2,RP,RP1).
recursive(T1,T2,TPs,RP,RP2,K,K1) :-
        ub1(T1,T2,T,D,TPs,K,K1),
        app(D,TPs,TPs1),
        normalise_defs(TPs1,T2,T,TPs2),
        replace_procs(TPs2,RP,RP1),
        normalise_defs(RP1,T1,T,RP2).

replace_procs([],R,R).
replace_procs([P|Ps],R,R1) :-
	replace_proc(P,R,R2),
	replace_procs(Ps,R2,R1).

replace_proc(proc(T1/1,Cls),[proc(T1/1,_)|R],[proc(T1/1,Cls)|R]) :-
	!.
replace_proc(P,[P1|R],[P1|R1]) :-
	replace_proc(P,R,R1).
replace_proc(P,[],[P]).


less_functors(T1,T2,R) :-
	\+ morefunctors(T1,T2,R).

morefunctors(T1,T2,R) :-
	headfunctor(F,T1,R),
	\+ headfunctor(F,T2,R).

same_functor_structure(1,T1,T2,RP) :-
	\+ morefunctors(T1,T2,RP),
	\+ morefunctors(T2,T1,RP).

same_functor_structure(2,T1,T2,R) :-
	\+ diffstructure(T1,T2,R),
	\+ morefunctors(T2,T1,R).

diffstructure(T1,T2,R) :-
	typeclause(T1/1,R,(H :- B)),
	H =.. [T1,X],
	functor(X,F,N),
	\+ corresponding_clause(T2,F/N,B,R).

corresponding_clause(T,F/N,B,R) :-
	typeclause(T/1,R,(H :- B1)),
	H =.. [T,X],
	functor(X,F,N),
	same_body_structure(B,B1,R).

same_body_structure(true,true,_).
same_body_structure((any(_),Bs),(any(_),Bs1),R) :-
	!,
	same_body_structure(Bs,Bs1,R).
same_body_structure((B,Bs),(B1,Bs1),R) :-
	!,
	functor(B,T,1),
	functor(B1,T1,1),
	T \== any,
	T1 \== any,
	same_functor_structure(1,T,T1,R),
	same_body_structure(Bs,Bs1,R).
same_body_structure(B,B1,R) :-
	functor(B,T,1),
	functor(B1,T1,1),
	T \== any,
	T1 \== any,
	same_functor_structure(1,T,T1,R).

headfunctor(F/N,T,Defs) :-
        typeclause(T/1,Defs,(H :- _)),
        H =.. [T,X],
        functor(X,F,N).


	
depends(T1,T2,R1,[_|P]) :-
	nested_memb(proc(T1/1,_),P),
	iterate_defs([T1],[T1|S],R1,[],P),
	makelist(S),
	memb2(T2,S).

makelist([]) :-
	!.
makelist([_|X]) :-
	makelist(X).

body_t([(_ :- B)|_],T3) :-
	lit(A,B),
	functor(A,T3,1).
body_t([_|Cls],T3) :-
	body_t(Cls,T3).

	

tp_alpha([],_,_,[],[],[],K,K).
tp_alpha([P|Ps],Nofix,R,[PR|PRs],[UP|R1],[Ch|Chs],K,K3) :-
	user_clauses(P,Cls),
	infer_proc(Cls,Nofix,R,As,K,K1),
	ub_proc(As,PR,PR,UP,Ch,K1,K2),
	tp_alpha(Ps,Nofix,R,PRs,R1,Chs,K2,K3).

infer_proc([(_ :- B)|Cls],Nofix,R,As,K,K1) :- 	% if no predicate in the body
	nochange(B,Nofix),			% changed, skip clause
	!,
	infer_proc(Cls,Nofix,R,As,K,K1).
infer_proc([(H :- B)|Cls],Nofix,R,[A2|As],K,K4) :-
	tsolve(B,R,T,Bdefs),
	unfold_body(T,Bdefs,T1,Adefs,K,K1),
	intersect_body(T1,Bdefs,T2,S,K1,K2),
	!,
	headtype(H,T2,_,A,K2,K3), 	
	app(Adefs,Bdefs,ABdefs),
	app(S,ABdefs,A1),
	app(A,A1,A2),
	infer_proc(Cls,Nofix,R,As,K3,K4).
infer_proc([_|Cls],Nofix,R,As,K,K1) :-		% if no solution in R.
	infer_proc(Cls,Nofix,R,As,K,K1).
infer_proc([],_,_,[],K,K).

nochange(_,[[]]) :-  	% everything handled on 1st iteration
	!,
	fail.
nochange(true,_) :-
	!.
nochange(B,Nofix) :-
	preds(B,Ps),
	\+ commonpred(Ps,Nofix).

preds(true,[]) :-
	!.
preds((B,Bs), [P/N|Ps]) :-
	!,
	functor(B,P,N),
        preds(Bs,Ps).
preds(B,[P/N]) :-
	functor(B,P,N).

commonpred(X,Y) :-
	memb2(Z,X),
	memb2(Z,Y).
	
tsolve(true,_,true,[]) :-
	!.
tsolve((B,Bs),R,T,Ds1) :-
	!,
	tsolve(B,R,T1,D),
	tsolve(Bs,R,T2,Ds),
	app(D,Ds,Ds1),
	joingoals(T1,T2,T).
tsolve(X=Y,_,true,[]) :-
        !,
	X=Y.
tsolve(B,_,Bs,[]) :-
        sp_builtin(B),
        !,
        builtin_type(B,Bs).
tsolve(B,R,T1,Bdef) :-
	functor(B,S,N),
	headclause(S/N,R,(H :- T),Bdef),
	melt((H :- T),(H1 :- T1)),
	H1 =.. [_,B].

unfold_body(true,_,true,[],K,K) :-
	!.
unfold_body((B,Bs),R,C,A1,K,K2) :-
	!,
	unfold_body(B,R,C1,A2,K,K1),
	unfold_body(Bs,R,C2,A3,K1,K2),
	app(A2,A3,A1),
	joingoals(C1,C2,C).
unfold_body(B,_,B,[],K,K) :-
	B =.. [_,X],
	var(X),
	!.
unfold_body(B,_,Bs,[],K,K) :-
	functor(B,any,1),
	!,
	any_term(B,Bs).
unfold_body(B,R,B2,A1,K,K1) :-
	B =.. [T,_],
	typeclause(T/1,R,(H :- B1)),
	matches((H :- B1),(B :- B3)),
	!,
	unfold_body(B3,R,B2,A1,K,K1).
unfold_body(B,R,B2,A1,K,K1) :-
	B =.. [T,_],
	def(T/1,P,R),
	alpha_match(B,P,B2,A1,K,K1).

alpha_match(B,P,B2,[proc(T/1,A1)],K,K1) :-
	extract_alpha_clauses(P,[A|As]),
	newname(K,T),
	K1 is K+1,
	alpha_def([A|As],T,A1),
	alpha_term(B,T,B2).

extract_alpha_clauses([],[]).
extract_alpha_clauses([(H :- true)|P],[(H :- true)|Q]) :-
	H =.. [_,X],
	alpha_const(X),
	!,
	extract_alpha_clauses(P,Q).
extract_alpha_clauses([_|P],Q) :-
	extract_alpha_clauses(P,Q).

alpha_const(X) :-
        name(X,Xn),
        name(alpha,A),
        app(A,_,Xn).

/*
alpha_def([],_,[]).
alpha_def([(H :- true)|As],T,[(H1:- true)|Bs]) :-
	H =.. [_,Alpha],
	H1 =.. [T,Alpha],
	alpha_def(As,T,Bs).
*/
alpha_def(_,T,[(H :- true)]) :-
	H =.. [T,alpha].

alpha_term(B,T,Bs) :-
	vars(B,Vs),
	alpha_vars(Vs,T,Bs).

alpha_vars([],_,true) :-
	!.
alpha_vars([X],T,B) :-
	!,
	B =.. [T,X].
alpha_vars([X|Xs],T,(B,Bs)) :-
	B =.. [T,X],
	alpha_vars(Xs,T,Bs).

matches((H :- B1),(B :- B3)) :-
	melt((H :- B1),(B :- B3)),
	!.
matches((H :- true),(B :- true)) :-
	H =.. [_,K],
	basic_type(K),
	B =.. [_,K1],
	atomic(K1),
	type_check(K1,K).

any_term(B,Bs) :-
	vars(B,Vs),
	any_vars(Vs,Bs).

any_vars([],true) :-
	!.
any_vars([X],any(X)) :-
	!.
any_vars([X|Xs],(any(X),Bs)) :-
	any_vars(Xs,Bs).
	
/*
any_number(B,Bs) :-
	vars(B,Vs),
	any_numbers(Vs,Bs).

any_list(list(X),list(X)) :-
	var(X),
	!.
any_list(list([]),true) :-
any_list(list([X|Y]),(any(X),any(Y))).

any_numbers(numeric(X),numeric(X)) :-
	var(X),
	!.
any_numbers(numeric(X),true) :-
	number(X),
	!.
any_number(numeric(X+Y),(numeric(X),numeric(Y))).
any_number(numeric(X*Y),(numeric(X),numeric(Y))).
any_number(numeric(X-Y),(numeric(X),numeric(Y))).
any_number(numeric(X/Y),(numeric(X),numeric(Y))).
any_number(numeric(-X),numeric(X)).
any_number(numeric(+X),numeric(X)).
any_number(numeric(X mod Y),(numeric(X),numeric(Y))).

*/
	
intersect_body(true,_,true,[],K,K) :-
	!.
intersect_body((B,Bs),R,C,I,K,K2) :-
	reoccurs(B,Bs,B1,Bs1),
	!,
	B =.. [T1,X],
	B1 =.. [T2,X],
	intersect(T1,T2,T3,P,R,K,K1),
	B2 =.. [T3,X],
	app(P,R,PR),
        intersect_body((B2,Bs1),PR,C,J,K1,K2),
	app(P,J,I).
intersect_body((B,Bs),R,(B,Bs1),I,K,K1) :-
	!,
	intersect_body(Bs,R,Bs1,I,K,K1).
intersect_body(B,_,B,[],K,K).
	
	
reoccurs(B,(B1,Bs),B1,Bs) :-
	B =.. [_,X],
	B1 =.. [_,Y],
	X == Y,
	!.
reoccurs(B,(B2,Bs),B1,(B2,Bs1)) :-
	!,
	reoccurs(B,Bs,B1,Bs1).
reoccurs(B,B1,B1,true) :-
	B =.. [_,X],
	B1 =.. [_,Y],
	X == Y,
	!.
	
headtype(T, Body,P,[proc(P/1,[(S :- B)])|Argtypes],K,K2) :- 
	functor(T,F,N),
	functor(X,F,N),
	newname(K,P),
	S =.. [P,X],
	canonical(S),
	K1 is K+1,
	argtypes(T,Body,X,B,1,N,K1,K2,Argtypes).
	
termtype(T, Body,[proc(P/1,[(S :- B)])|Argtypes],K,K2) :- 
	functor(T,F,N),
	functor(X,F,N),
	newname(K,P),
	S =.. [P,X],
	canonical(S),
	K1 is K+1,
	argtypes(T,Body,X,B,1,N,K1,K2,Argtypes).

argtypes(_,_,_,true,_,0,K,K,[]).
argtypes(T,Body,X,B,N,N,K,K1,Argtypes) :-
	argtype(T,Body,X,B,N,K,K1,Argtypes).
argtypes(T,Body,X,(B,Bs),M,N,K,K2,Argtypes) :-
	M < N,
	argtype(T,Body,X,B,M,K,K1,Argtype),
	M1 is M+1,
	argtypes(T,Body,X,Bs,M1,N,K1,K2,Argtypes1),
	app(Argtype,Argtypes1,Argtypes). 
	
argtype(T,Body,X,B,M,K,K,[]) :-
	arg(M,T,Xm),
	var(Xm),
	!,
	arg(M,X,Ym),
	giventype(Xm,Body,R),
	B =.. [R,Ym].
argtype(T,Body,X,B,M,K,K1,Types) :-
        arg(M,T,Xm),
        arg(M,X,Ym),
        termtype(Xm,Body,Types,K,K1),
        newname(K,P),
        B =.. [P,Ym].


builtin_type_functor(X,numeric) :-
	number(X),	
	!.
builtin_type_functor(_+_,numeric).
builtin_type_functor(_*_,numeric).
builtin_type_functor(_/_,numeric).
builtin_type_functor(_-_,numeric).
builtin_type_functor(-_,numeric).
builtin_type_functor(+_,numeric).
builtin_type_functor(_ mod _,numeric).
builtin_type_functor([],list).
builtin_type_functor([_|_],list).

builtin_type(X is Y, (numeric(X),numexpr(Y))) :-
	!.
builtin_type(X < Y, (numexpr(X),numexpr(Y))) :-
	!.
builtin_type(X > Y, (numexpr(X),numexpr(Y))) :-
	!.
builtin_type(X =< Y, (numexpr(X),numexpr(Y))) :-
	!.
builtin_type(X >= Y, (numexpr(X),numexpr(Y))) :-
	!.
builtin_type(number(X), numeric(X)) :-
	!.
builtin_type(integer(X), numeric(X)) :-
	!.
builtin_type(B, Bs) :-
	any_term(B,Bs).
	
giventype(X,(B,_),T) :-
	B =.. [T,Y],
	X == Y,
	!.
giventype(X,(_,Bs),T) :-
	!,
	giventype(X,Bs,T).
giventype(X,B,T) :-
	B =.. [T,Y],
	X == Y,
	!.
giventype(_,_,any).
	
newname(K,P) :-
	name(K,Kn),
	name(t,T),
	app(T,Kn,Tn),
	name(P,Tn).

check_fix([],[],[]).
check_fix([P|Ps],[change|Cs],[P|Chs]) :-
        check_fix(Ps,Cs,Chs).
check_fix([_|Ps],[nochange|Cs],Chs) :-
        check_fix(Ps,Cs,Chs).


check_fix([],[],[],[]).
check_fix([_|Ps],[RP|R],[RP1|S],Qs) :-
        fixpoint(RP,RP1),
	!,
        check_fix(Ps,R,S,Qs).
check_fix([P|Ps],[_|R],[_|S],[P|Qs]) :-
        check_fix(Ps,R,S,Qs).

fixpoint([],[]) :-
	!.
fixpoint(R1,R2) :-
	toptype(R1,T1),
	toptype(R2,T2),
	subtype(T2,T1,[R1|R2]).
	
toptype([proc(T1/1,_)|_],T1).

ub_proc([],_,R1,R1,nochange,K,K) :-  
        !.
ub_proc([P],R,[proc(TP/1,_)|_],D,nochange,K,K) :-  % ub with current def
	toptype(P,Q),
	app(P,R,RP),
        subtype(Q,TP,RP),
	!,
	rdefs(TP,D,RP).
ub_proc([P],R,[proc(TP/1,_)|_],D,change,K,K1) :-  % upper bound with current def
	!,
	toptype(P,Q),
	app(P,R,RP),
        ub1(TP,Q,Q1,D1,RP,K,K1),
        app(D1,RP,RPD),
        rdefs(Q1,D,RPD).
ub_proc([P],R,_,D,change,K,K) :-  
	!,
	toptype(P,Q),
	app(P,R,RP),
        rdefs(Q,D,RP).
ub_proc([P1,P2|Ps],R,R1,U,Ch,K,K2) :-
	toptype(P1,T1),
	toptype(P2,T2),
	app(P2,R,P3),
        app(P1,P3,RP),
        subtype(T1,T2,RP),
        !,
        app(P2,R,PR2),
        ub_proc([P2|Ps],PR2,R1,U,Ch,K,K2).
ub_proc([P1,P2|Ps],R,R1,U,Ch,K,K2) :-
	toptype(P1,T1),
	toptype(P2,T2),
	app(P2,R,P3),
        app(P1,P3,RP),
        subtype(T2,T1,RP),
        !,
        app(P1,R,PR1),
        ub_proc([P1|Ps],PR1,R1,U,Ch,K,K2).
ub_proc([P1,P2|Ps],R,R1,U,Ch,K,K2) :-
	toptype(P1,T1),
	toptype(P2,T2),
	app(P2,R,P3),
        app(P1,P3,RP),
        ub1(T1,T2,_,D,RP,K,K1),
        app(D,RP,DRP),
        ub_proc([D|Ps],DRP,R1,U,Ch,K1,K2).

headclause(P/N,R,C,Xdef) :-
	functor(X,P,N),
	hclause(X,R,C,Xdef).

hclause(H,[[proc(T/1,[(TH :- B)])|Hdef]|_],(TH :- B),Hdef) :-
	TH =.. [T,H],
	!.
hclause(P,[_|Ps],Cl,Def) :-
	hclause(P,Ps,Cl,Def).


nested_memb(X,[X|_]).
nested_memb(X,[Y|_]) :-
	nested_memb(X,Y).
nested_memb(X,[_|Z]) :-
	nested_memb(X,Z).


typeclause(T/N,P,Cl) :-
	def(T/N,Q,P),
	!,
	memb2(Cl,Q).


merge_procs([],R,R).
merge_procs([P|Ps],R1,R2) :-
	insert_proc(P,R1,R3),
	merge_procs(Ps,R3,R2).

insert_proc(P,R,R) :-
	procmemb(P,R),
	!.
insert_proc(P,R,[P|R]).

procmemb(proc(F/N,_),[proc(F/N,_)|_]).
procmemb(P,[_|Ps]) :-
	procmemb(P,Ps).

:- dynamic traceon/0.

writestate(R) :-
	traceon,
	write(R),
	read(_),
	!,
	nl.
writestate(_) :-
	write('*'),
	sp_ttyflush.

lit(B, (B,_)).
lit(B, (_,Bs)) :-
        !,
        lit(B,Bs).
lit(B,B).

display_fix(Cls) :-
        nl,write('Output file: '),
        read(F),
        reg_filewrite(F,Cls).

reg_filewrite(tty,Cls) :-
        !,
	nl,write('Regular Approximation Program:'),nl,nl,
        displayqTypes(Cls),
	nl,nl.
reg_filewrite(F,Cls) :-
        tell(F),
        displayqTypes(Cls),
        told.

/* write the program to current output file  */

displayqTypes([[]|Ps]) :-
	displayqTypes(Ps).
displayqTypes([[Q|Qs]|Ps]) :-
	displayqPreddef(Q),
	displayqTypes(Qs),
	displayqTypes(Ps).
displayqTypes([proc(_,Qs)|Ps]) :-
	displayqRClauses(Qs),
	displayqTypes(Ps).
displayqTypes([]).

displayqPreddef(proc(_,[(H :- B)])) :-
	H =.. [_,A],
	displayqRClauses([(A :- B)]).


show_useless_clauses(P,F,K,K1) :-
        new_line,
        new_line,
        write('Find useless clauses (y/n)? '),
        getchar(YN),
        show_useless_clauses(YN,P,F,K,K1).

show_useless_clauses(y,P,F,K,K1) :-
        new_line,
        write('The following clauses are useless:'),
        new_line,
        new_line,
        !,
        useless_clause_check(P,F,K,K1,Us,[]),
        new_line,
        write('The following predicates are useless:'),
        new_line,
        write(Us).

show_useless_clauses(_,_,_,K,K).

useless_clause_check([],_,K,K,Us,Us).
useless_clause_check([P|Ps],R,K,K2,Us,Us2) :-
        user_clauses(P,Cls),
        check_bodies(P,1,Cls,R,K,K1,U),
        useless_pred(P,U,Us,Us1),
        useless_clause_check(Ps,R,K1,K2,Us1,Us2).

check_bodies(P,J,[(_ :- true)|Cls],R,K,K1,[J|U]) :-
        !,
        J1 is J+1,
        check_bodies(P,J1,Cls,R,K,K1,U).
check_bodies(P,J,[(_ :- B)|Cls],R,K,K2,[J|U]) :-
        tsolve(B,R,T,Bdefs),
        unfold_body(T,Bdefs,T1),
        intersect_body(T1,Bdefs,_,_,K,K1),
        !,
        J1 is J+1,
        check_bodies(P,J1,Cls,R,K1,K2,U).
check_bodies(P,J,[(H :- B)|Cls],R,K,K1,U) :-          % if no solution in R.
        canonical((H :- B)),
%       write((H :- B)),
        write(P),
        write(': Clause '),
        write(J),
        new_line,
        J1 is J+1,
        check_bodies(P,J1,Cls,R,K,K1,U).
check_bodies(_,_,[],_,K,K,[]).

useless_pred(P,[],[P|Us],Us) :-
        !.
useless_pred(_,_,Us,Us).

