% JPG - 23.1.2000

tp_fix(P) :-
	time(tp_fix1(P,F)),
	display_fix(F).

tp_fix1(P,F) :-
	strong_connected_components(P,G),	% compute the dependency ordering
	nl,
	tp_fix_groups(G,[],F,[]),
	show_useless_clauses(P,F).

tp_fix_groups([],F,F,_) :-
	!.
tp_fix_groups([(R,G)|Gs],F,F2,I0) :-
	nonrec(R),
	write(G),nl,
	!,
	init(G,R0),
	tp_alpha(G,G,F,R0,R2,_,I0,I1),  % no need to iterate if non-rec.
	app(F,R2,F1),
	tp_fix_groups(Gs,F1,F2,I1).
tp_fix_groups([(_,G)|Gs],F,F2,I0) :-
	write(G),nl,
	init(G,R0),
	tp_iterate(G,G,F,R0,R2,I0,I1),
	app(R2,F,F1),
	tp_fix_groups(Gs,F1,F2,I1).

nonrec(non_recursive).

recursive_pred(X) :-
	memb2(P,X),
	recursive(P),
	!.

init([_|Ps],[[]|R]) :-
	init(Ps,R).
init([],[]).

tp_iterate([],_,_,R,R,I,I) :-
	!.
tp_iterate(Nofix,Ps,Acc,R1,R,I0,I2) :-
	writestate(R1),
	app(R1,Acc,D),
	tp_alpha(Ps,Nofix,D,R1,R2,Chs,I0,I1),
	countdefs(R2),
	check_fix(Ps,Chs,Fix),
	!,
	tp_iterate(Fix,Ps,Acc,R2,R,I1,I2).

countdefs([R|Rs]) :-
	sp_length(R,N),write(N),write(' '),
	!,
	countdefs(Rs).
countdefs([]) :-
	nl.

makelist([]) :-
	!.
makelist([_|X]) :-
	makelist(X).

body_t([(_ :- B)|_],T3) :-
	lit(A,B),
	functor(A,T3,1).
body_t([_|Cls],T3) :-
	body_t(Cls,T3).

	
tp_alpha([],_,_,[],[],[],I,I).
tp_alpha([P|Ps],Nofix,D,[PDefs|Ps],[PDefs1|Ps1],[Ch|Chs],I0,I1) :-
	user_clauses(P,Cls),
	infer_proc(Cls,Nofix,D,I0,I1,As),
	update(PDefs,As,Pdefs1,Ch),
	tp_alpha(Ps,Nofix,R,Ps,Ps1,Chs,I1,I2).

infer_proc([(_ :- B)|Cls],Nofix,R,I,I1,As) :- 	% if no predicate in the body
	nochange(B,Nofix),				% changed, skip clause
	!,
	infer_proc(Cls,Nofix,R,I,I1,As).
infer_proc([(H :- B)|Cls],Nofix,R,I,I2,As1) :-
	normal_head((H :- B),R,Cls),
	unfold_body(Cls,R,[],Us),
	intersect_check(Us,Us1,R,I,I1),	
	infer_proc(Cls,Nofix,R,I1,I2,As),
	app(Us1,As,As1).
infer_proc([],_,_,I,I,[]).

nochange(_,[[]]) :-  	% everything handled on 1st iteration
	!,
	fail.
nochange(true,_) :-
	!.
nochange(B,Nofix) :-
	preds(B,Ps),
	\+ commonpred(Ps,Nofix).

preds(true,[]) :-
	!.
preds((B,Bs), [P/N|Ps]) :-
	!,
	functor(B,P,N),
	preds(Bs,Ps).
preds(B,[P/N]) :-
	functor(B,P,N).

commonpred(X,Y) :-
	memb2(Z,X),
	memb2(Z,Y).
	
normal_head((H :- B), D, [(H,[],B)|E],E) :-
	H =.. [_,T],
	nonvar(T),
	!.
normal_head((H :- B), D, E0,E1) :-
	resolvents(H,[],B,D,Rs),
	normal_head_each(Rs,D,E0,E1).

normal_head_each([],_,E,E).
normal_head_each([(H,[],B)|Rs],D,E0,E2) :-
	normal_head((H :- B), D, E0,E1),
	normal_head_each(Rs,D,E1,E2).

unfold_body([],_,E,E).
unfold_body([(H,Pre,Bs)|F],D,E0,E2) :-
	resolvents(H,Pre,Bs,D,Rs),
	unfold_body(Rs,D,E0,E1),
	unfold_body(F,D,E1,E2).

resolvents(H,Pre,true,_,[(H :- [],R)]) :-
	!,
	revgoal(Pre,R).
resolvents(H,Pre,(B,Bs),D,Rs) :-
	B =.. [_,X],
	var(X),
	!,
	resolvents(H,[B|Pre],Bs,D,Rs).
resolvents(H,Pre,(B,Bs),D,Rs) :-
	!,
	functor(B,F,1),
	defs(F,D,FDefs),
	resolvents1(FDefs,(H,Pre,B,Bs),Rs).
resolvents(H,Pre,B,D,Rs) :-
	B =.. [_,X],
	var(X),
	!,
	revgoal([B|Pre],Rs).
resolvents(H,Pre,B,D,Rs) :-
	!,
	functor(B,F,1),
	defs(F,D,FDefs),
	resolvents1(FDefs,(H,Pre,B,true),Rs).

resolvents1([],_,_,_,_,[]).
resolvents1([(A :- R)|FDefs],G,[(H1,Pre1,Bs2)|Rs]):-
	copyterm(G,(H1,Pre1,B1,Bs1)),
	melt((A,R),(H1,R1)),
	!,
	joingoals(R1,Bs1,Bs2),
	resolvents1(FDefs,G,Rs).

check_intersect([],[],_,I,I).
check_intersect([(H :- [],B)|Rs],[(H :- B2)|Rs1],D,I0,I2) :-
	sort_body(B,B1),
	each_non_empty(B1,D,I0,I1,[]),
	!,
	headvars(H,Ws),
	project_body(Ws,B1,B2),
	check_intersect(Rs,Rs1,D,I1,I2).
check_intersect([_|Rs],Rs1,D,I0,I1) :-
	check_intersect(Rs,Rs1,D,I0,I1).

each_non_empty([],_,I,I,_).
each_non_empty([(_,Plist)|B],_,I,I,_) :-
	subsumed(Plist,I),
	!.
each_non_empty([(_,Plist)|B],D,I0,I4,Anc) :-
	\+ member(Plist,Anc),
	one_step_expand(Plist,Qlist),
	each_non_empty(Qlist,D,I0,I1,[Plist|Anc]),
	insertset(Plist,I1,I2),
	delete_subsumed(I2,I3),
	each_non_empty(B,D,I3,I4,Anc).
	
	
any_term(B,Bs) :-
	vars(B,Vs),
	any_vars(Vs,Bs).

any_vars([],true) :-
	!.
any_vars([X],any(X)) :-
	!.
any_vars([X|Xs],(any(X),Bs)) :-
	any_vars(Xs,Bs).
		
	
reoccurs(B,(B1,Bs),B1,Bs) :-
	B =.. [_,X],
	B1 =.. [_,Y],
	X == Y,
	!.
reoccurs(B,(B2,Bs),B1,(B2,Bs1)) :-
	!,
	reoccurs(B,Bs,B1,Bs1).
reoccurs(B,B1,B1,true) :-
	B =.. [_,X],
	B1 =.. [_,Y],
	X == Y,
	!.
	
headtype(T, Body,P,[proc(P/1,[(S :- B)])|Argtypes],K,K2) :- 
	functor(T,F,N),
	functor(X,F,N),
	newname(K,P),
	S =.. [P,X],
	canonical(S),
	K1 is K+1,
	argtypes(T,Body,X,B,1,N,K1,K2,Argtypes).
	
termtype(T, Body,[proc(P/1,[(S :- B)])|Argtypes],K,K2) :- 
	functor(T,F,N),
	functor(X,F,N),
	newname(K,P),
	S =.. [P,X],
	canonical(S),
	K1 is K+1,
	argtypes(T,Body,X,B,1,N,K1,K2,Argtypes).

argtypes(_,_,_,true,_,0,K,K,[]).
argtypes(T,Body,X,B,N,N,K,K1,Argtypes) :-
	argtype(T,Body,X,B,N,K,K1,Argtypes).
argtypes(T,Body,X,(B,Bs),M,N,K,K2,Argtypes) :-
	M < N,
	argtype(T,Body,X,B,M,K,K1,Argtype),
	M1 is M+1,
	argtypes(T,Body,X,Bs,M1,N,K1,K2,Argtypes1),
	app(Argtype,Argtypes1,Argtypes). 
	
argtype(T,Body,X,B,M,K,K,[]) :-
	arg(M,T,Xm),
	var(Xm),
	!,
	arg(M,X,Ym),
	giventype(Xm,Body,R),
	B =.. [R,Ym].
argtype(T,Body,X,B,M,K,K1,Types) :-
	arg(M,T,Xm),
	arg(M,X,Ym),
	termtype(Xm,Body,Types,K,K1),
	newname(K,P),
	B =.. [P,Ym].

giventype(X,(B,_),T) :-
	B =.. [T,Y],
	X == Y,
	!.
giventype(X,(_,Bs),T) :-
	!,
	giventype(X,Bs,T).
giventype(X,B,T) :-
	B =.. [T,Y],
	X == Y,
	!.
giventype(_,_,any).
	
newname(K,P) :-
	name(K,Kn),
	name(t,T),
	app(T,Kn,Tn),
	name(P,Tn).

check_fix([],[],[]).
check_fix([P|Ps],[change|Cs],[P|Chs]) :-
        check_fix(Ps,Cs,Chs).
check_fix([_|Ps],[nochange|Cs],Chs) :-
        check_fix(Ps,Cs,Chs).

toptype([proc(T1/1,_)|_],T1).


headclause(P/N,R,C,Xdef) :-
	functor(X,P,N),
	hclause(X,R,C,Xdef).

hclause(H,[[proc(T/1,[(TH :- B)])|Hdef]|_],(TH :- B),Hdef) :-
	TH =.. [T,H],
	!.
hclause(P,[_|Ps],Cl,Def) :-
	hclause(P,Ps,Cl,Def).



nested_memb(X,[X|_]).
nested_memb(X,[Y|_]) :-
	nested_memb(X,Y).
nested_memb(X,[_|Z]) :-
	nested_memb(X,Z).


typeclause(T/N,P,Cl) :-
	def(T/N,Q,P),
	!,
	memb2(Cl,Q).
	


merge_procs([],R,R).
merge_procs([P|Ps],R1,R2) :-
	insert_proc(P,R1,R3),
	merge_procs(Ps,R3,R2).

insert_proc(P,R,R) :-
	procmemb(P,R),
	!.
insert_proc(P,R,[P|R]).

procmemb(proc(F/N,_),[proc(F/N,_)|_]).
procmemb(P,[_|Ps]) :-
	procmemb(P,Ps).

:- dynamic traceon/0.

writestate(R) :-
	traceon,
	write(R),
	read(_),
	!,
	nl.
writestate(_) :-
	write('*'),
	sp_ttyflush.

lit(B, (B,_)).
lit(B, (_,Bs)) :-
        !,
        lit(B,Bs).
lit(B,B).

display_fix(Cls) :-
        nl,write('Enter filename: '),
        read(F),
        reg_filewrite(F,Cls).

reg_filewrite(tty,Cls) :-
        !,
	nl,write('Regular Approximation Program:'),nl,nl,
        displayqTypes(Cls),
	nl,nl.
reg_filewrite(F,Cls) :-
        tell(F),
        displayqTypes(Cls),
        told.

/* write the program to current output file  */

displayqTypes([[]|Ps]) :-
	displayqTypes(Ps).
displayqTypes([[Q|Qs]|Ps]) :-
	displayqPreddef(Q),
	displayqTypes(Qs),
	displayqTypes(Ps).
displayqTypes([proc(_,Qs)|Ps]) :-
	displayqRClauses(Qs),
	displayqTypes(Ps).
displayqTypes([]).

displayqPreddef(proc(_,[(H :- B)])) :-
	H =.. [_,A],
	displayqRClauses([(A :- B)]).


show_useless_clauses(P,F,K,K1) :-
	new_line,
	new_line,
	write('Find useless clauses (y/n)? '),
	getchar(YN),
	show_useless_clauses(YN,P,F,K,K1).

show_useless_clauses(y,P,F,K,K1) :-
	new_line,
	write('The following clauses are useless:'),
	new_line,
	new_line,
	!,
	useless_clause_check(P,F,K,K1,Us,[]),
	new_line,
	write('The following predicates are useless:'),
	new_line,
	write(Us).

show_useless_clauses(_,_,_,K,K).

useless_clause_check([],_,K,K,Us,Us).
useless_clause_check([P|Ps],R,K,K2,Us,Us2) :-
        user_clauses(P,Cls),
        check_bodies(P,1,Cls,R,K,K1,U),
	useless_pred(P,U,Us,Us1),
        useless_clause_check(Ps,R,K1,K2,Us1,Us2).

check_bodies(P,J,[(_ :- true)|Cls],R,K,K1,[J|U]) :-   
        !,
	J1 is J+1,
        check_bodies(P,J1,Cls,R,K,K1,U).
check_bodies(P,J,[(_ :- B)|Cls],R,K,K2,[J|U]) :-
        tsolve(B,R,T,Bdefs),
        unfold_body(T,Bdefs,T1),
        intersect_body(T1,Bdefs,_,_,K,K1),
        !,
	J1 is J+1,
        check_bodies(P,J1,Cls,R,K1,K2,U).
check_bodies(P,J,[(H :- B)|Cls],R,K,K1,U) :-          % if no solution in R.
	canonical((H :- B)),
%	write((H :- B)),
	write(P),
	write(': Clause '), 
	write(J),
	new_line,
	J1 is J+1,
        check_bodies(P,J1,Cls,R,K,K1,U).
check_bodies(_,_,[],_,K,K,[]).

useless_pred(P,[],[P|Us],Us) :-
	!.
useless_pred(_,_,Us,Us).
