% JPG - 14.2.92
% Compute a regular approximation of a given program.
% For a description of the algorithm,
% see Gallagher and de Waal:  "Regular Approximations of
% Logic Programs and Their Uses";  CSTR-92-06.

%:- setdebug.
 
%:- alldynamic.

tp_fix(P) :-
	time(tp_fix1(P,F)),
	display_fix(F).

tp_fix1(P,F) :-
	strong_connected_components(P,G),% compute the dependency ordering
	nl,
	tp_fix_groups(G,[],F,10,K),
	show_useless_clauses(P,F,K,_).

tp_fix_groups([],F,F,K,K) :-
	!.
tp_fix_groups([(R,G)|Gs],F,F2,K,K2) :-
	nonrec(R),
	write(G),nl,
	!,
	init(G,R0),
	tp_alpha(G,[[]],F,R0,R2,_,K,K1),  % no need to iterate if non-rec.
	app(F,R2,F1),
	tp_fix_groups(Gs,F1,F2,K1,K2).
tp_fix_groups([(_,G)|Gs],F,F2,K,K2) :-
	write(G),nl,
	init(G,R0),
	tp_iterate([[]],G,F,R0,R2,K,K1),
	app(F,R2,F1),
	tp_fix_groups(Gs,F1,F2,K1,K2).

nonrec(non_recursive).

recursive_pred(X) :-
	memb2(P,X),
	recursive(P),
	!.

init([_|Ps],[[]|R]) :-
	init(Ps,R).
init([],[]).

tp_iterate([],_,_,R,R,K,K) :-
	!.
tp_iterate(Nofix,Ps,Acc,R1,R,K,K3) :-
	writestate(R1),
	app(R1,Acc,D),
	tp_alpha(Ps,Nofix,D,R1,R2,Chs,K,K1),
	fold(R2,R3,Chs,K1,K2),
	countdefs(R3),
	%check_fix(Ps,R1,R3,Fix),
	check_fix(Ps,Chs,Fix),
	!,
	tp_iterate(Fix,Ps,Acc,R3,R,K2,K3).

countdefs([R|Rs]) :-
	sp_length(R,N),write(N),write(' '),
	!,
	countdefs(Rs).
countdefs([]) :-
	nl.

fold([],[],[],K,K).
fold([RP|R1],[RP|R3],[nochange|Chs],K,K2) :-
	!,
	fold(R1,R3,Chs,K,K2).
fold([RP|R1],[RP2|R3],[change|Chs],K,K2) :-
        toptype(RP,T),
        rdefs(T,RP1,RP),
        foldp(RP1,RP2,K,K1),
        fold(R1,R3,Chs,K1,K2).

	
%:- dynamic level/1.

level(1).

foldp(RP,RP3,K,K2) :-
	depends(T1,T2,TPs,RP),
	level(Level),
	same_functor_structure(Level,T2,T1,TPs),
	recursive(T1,T2,TPs,RP,RP1,K,K1),
	!,
	toptype(RP1,T),
	rdefs(T,RP2,RP1),
	foldp(RP2,RP3,K1,K2).
foldp(R,R,K,K).

%recursive(T1,T2,TPs,RP,RP1,K,K) :-
%        subtype(T2,T1,TPs),
%        !,
%        normalise_defs(TPs,T2,T1,TPs1),
%        replace_procs(TPs1,RP,RP1).
recursive(T1,T2,TPs,RP,RP2,K,K1) :-
        ub1(T1,T2,T,D,TPs,K,K1),
        app(TPs,D,TPs1),
        normalise_defs(TPs1,T2,T,TPs2),
        replace_procs(TPs2,RP,RP1),
        normalise_defs(RP1,T1,T,RP2).

replace_procs([],R,R).
replace_procs([P|Ps],R,R1) :-
	replace_proc(P,R,R2),
	replace_procs(Ps,R2,R1).

replace_proc(proc(T1/1,Cls),[proc(T1/1,_)|R],[proc(T1/1,Cls)|R]) :-
	!.
replace_proc(P,[P1|R],[P1|R1]) :-
	replace_proc(P,R,R1).
replace_proc(P,[],[P]).


less_functors(T1,T2,R) :-
	\+ morefunctors(T1,T2,R).

morefunctors(T1,T2,R) :-
	headfunctor(F,T1,R),
	\+ headfunctor(F,T2,R).

same_functor_structure(1,T1,T2,RP) :-
	\+ morefunctors(T1,T2,RP),
	\+ morefunctors(T2,T1,RP).

same_functor_structure(2,T1,T2,R) :-
	\+ diffstructure(T1,T2,R),
	\+ morefunctors(T2,T1,R).

diffstructure(T1,T2,R) :-
	typeclause(T1/1,R,(H :- B)),
	H =.. [T1,X],
	functor(X,F,N),
	\+ corresponding_clause(T2,F/N,B,R).

corresponding_clause(T,F/N,B,R) :-
	typeclause(T/1,R,(H :- B1)),
	H =.. [T,X],
	functor(X,F,N),
	same_body_structure(B,B1,R).

same_body_structure(true,true,_).
same_body_structure((A,Bs),(A1,Bs1),R) :-
	anypred(A,Any),
	anypred(A1,Any),
	!,
	same_body_structure(Bs,Bs1,R).
same_body_structure((B,Bs),(B1,Bs1),R) :-
	!,
	\+ anypred(B,_),
	\+ anypred(B1,_),
	same_functor_structure(1,T,T1,R),
	same_body_structure(Bs,Bs1,R).
same_body_structure(B,B1,R) :-
	\+ anypred(B,_),
	\+ anypred(B1,_),
	same_functor_structure(1,T,T1,R).

headfunctor(F/N,T,Defs) :-
        typeclause(T/1,Defs,(H :- _)),
        H =.. [T,X],
        functor(X,F,N).


	
depends(T1,T2,R1,[_|P]) :-
	nested_memb(proc(T1/1,_),P),
	iterate_defs([T1],[T1|S],R1,[],P),
	makelist(S),
	memb2(T2,S).

makelist([]) :-
	!.
makelist([_|X]) :-
	makelist(X).

body_t([(_ :- B)|_],T3) :-
	lit(A,B),
	functor(A,T3,1).
body_t([_|Cls],T3) :-
	body_t(Cls,T3).

	

tp_alpha([],_,_,[],[],[],K,K).
tp_alpha([P|Ps],Nofix,R,[PR|PRs],[UP|R1],[Ch|Chs],K,K3) :-
	user_clauses(P,Cls),
	infer_proc(Cls,Nofix,R,As,K,K1),
	ub_proc(As,[As|PR],PR,UP,Ch,K1,K2),
	tp_alpha(Ps,Nofix,R,PRs,R1,Chs,K2,K3).

infer_proc([(_ :- B)|Cls],Nofix,R,As,K,K1) :- 	% if no predicate in the body
	nochange(B,Nofix),			% changed, skip clause
	!,
	infer_proc(Cls,Nofix,R,As,K,K1).
infer_proc([(H :- B)|Cls],Nofix,R,[A2|As],K,K3) :-
	tsolve(B,R,T,Bdefs),
	unfold_body(T,Bdefs,T1),
	intersect_body(T1,Bdefs,T2,S,K,K1),
	!,
	headtype(H,T2,_,A,K1,K2), 	
	app(S,Bdefs,A1),
	app(A,A1,A2),
	infer_proc(Cls,Nofix,R,As,K2,K3).
infer_proc([_|Cls],Nofix,R,As,K,K1) :-		% if no solution in R.
	infer_proc(Cls,Nofix,R,As,K,K1).
infer_proc([],_,_,[],K,K).

nochange(_,[[]]) :-  	% everything handled on 1st iteration
	!,
	fail.
nochange(true,_) :-
	!.
nochange(B,Nofix) :-
	preds(B,Ps),
	\+ commonpred(Ps,Nofix).

preds(true,[]) :-
	!.
preds((B,Bs), [P/N|Ps]) :-
	!,
	functor(B,P,N),
        preds(Bs,Ps).
preds(B,[P/N]) :-
	functor(B,P,N).

commonpred(X,Y) :-
	memb2(Z,X),
	memb2(Z,Y).
	
tsolve(true,_,true,[]) :-
	!.
tsolve((B,Bs),R,T,[D|Ds]) :-
	!,
	tsolve(B,R,T1,D),
	tsolve(Bs,R,T2,Ds),
	joingoals(T1,T2,T).
tsolve(X=Y,_,true,[]) :-
        !,
	X=Y.
tsolve(B,_,Bs,[]) :-
        sp_builtin(B),
        !,
        any_term(B,any,Bs).
tsolve(B,R,T1,Bdef) :-
	functor(B,S,N),
	headclause(S/N,R,(H :- T),Bdef),
	melt((H :- T),(H1 :- T1)),
	H1 =.. [_,B].

unfold_body(true,_,true) :-
	!.
unfold_body((B,Bs),R,C) :-
	!,
	unfold_body(B,R,C1),
	unfold_body(Bs,R,C2),
	joingoals(C1,C2,C).
unfold_body(B,_,B) :-
	B =.. [_,X],
	var(X),
	!.
unfold_body(B,_,Bs) :-
	anypred(B,Any),
	!,
	any_term(B,Any,Bs).
unfold_body(B,R,B2) :-
	B =.. [T,_],
	typeclause(T/1,R,(H :- B1)),
	melt((H :- B1),(B :- B3)),
	unfold_body(B3,R,B2).

anypred(B,Any) :-
	functor(B,Any,1),
	name(Any,[97,110,121|Num]),	% a = 97, n = 110, y = 121
	name(N,Num),
	integer(N).

any_term(B,Any,Bs) :-
	vars(B,Vs),
	any_vars(Vs,Any,Bs).

any_vars([],_,true) :-
	!.
any_vars([X],Any,A)) :-
	!,
	A =.. [Any,X].
any_vars([X|Xs],Any,(A,Bs)) :-
	A =.. [Any,X].
	any_vars(Xs,Any,Bs).
	
intersect_body(true,_,true,[],K,K) :-
	!.
intersect_body((B,Bs),R,C,I,K,K2) :-
	reoccurs(B,Bs,B1,Bs1),
	!,
	B =.. [T1,X],
	B1 =.. [T2,X],
	intersect(T1,T2,T3,P,R,K,K1),
	B2 =.. [T3,X],
        intersect_body((B2,Bs1),[P|R],C,J,K1,K2),
	app(P,J,I).
intersect_body((B,Bs),R,(B,Bs1),I,K,K1) :-
	!,
	intersect_body(Bs,R,Bs1,I,K,K1).
intersect_body(B,_,B,[],K,K).
	
	
reoccurs(B,(B1,Bs),B1,Bs) :-
	B =.. [_,X],
	B1 =.. [_,Y],
	X == Y,
	!.
reoccurs(B,(B2,Bs),B1,(B2,Bs1)) :-
	!,
	reoccurs(B,Bs,B1,Bs1).
reoccurs(B,B1,B1,true) :-
	B =.. [_,X],
	B1 =.. [_,Y],
	X == Y,
	!.
	
headtype(T, Body,P,[proc(P/1,[(S :- B)])|Argtypes],K,K2) :- 
	listvars(T,TVars),
	vars(Body,BVars),
	functor(T,F,N),
	functor(X,F,N),
	newname(K,P),
	S =.. [P,X],
	canonical(S),
	K1 is K+1,
	argtypes(T,Body,X,B,1,N,K1,K2,Argtypes,TVars,BVars).

termtype(T, Body,[proc(P/1,[(S :- B)])|Argtypes],K,K2,Tv,Bv) :- 
	functor(T,F,N),
	functor(X,F,N),
	newname(K,P),
	S =.. [P,X],
	canonical(S),
	K1 is K+1,
	argtypes(T,Body,X,B,1,N,K1,K2,Argtypes,Tv,Bv).

argtypes(_,_,_,true,_,0,K,K,[],_,_).
argtypes(T,Body,X,B,N,N,K,K1,Argtypes,Tv,Bv) :-
	argtype(T,Body,X,B,N,K,K1,Argtypes,Tv,Bv).
argtypes(T,Body,X,(B,Bs),M,N,K,K2,Argtypes,Tv,Bv) :-
	M < N,
	argtype(T,Body,X,B,M,K,K1,Argtype,Tv,Bv),
	M1 is M+1,
	argtypes(T,Body,X,Bs,M1,N,K1,K2,Argtypes1,Tv,Bv),
	app(Argtype,Argtypes1,Argtypes). 
	
argtype(T,Body,X,B,M,K,K,[],Tv,Bv) :-
	arg(M,T,Xm),
	var(Xm),
	!,
	arg(M,X,Ym),
	giventype(Xm,Body,R,Tv,Bv),
	B =.. [R,Ym].
argtype(T,Body,X,B,M,K,K1,Types,Tv,Bv) :-
	arg(M,T,Xm),
	arg(M,X,Ym),
	termtype(Xm,Body,Types,K,K1,Tv,Bv),
	newname(K,P),
	B =.. [P,Ym].

giventype(X,B,T,_,Bv) :-
	memb3(X,Bv),
	!,
	giventype(X,B,T).
giventype(X,_,sharevar,Tv,_) :-
	repeated(X,Tv),
	!.
giventype(_,_,freevar,_,_).

giventype(X,(B,_),T) :-
	B =.. [T,Y],
	X == Y,
	!.
giventype(X,(_,Bs),T) :-
	!,
	giventype(X,Bs,T).
giventype(X,B,T) :-
	B =.. [T,Y],
	X == Y,
	!.
	
repeated(X,[Y|Ys]) :-
	X == Y,
	!,
	memb3(X,Ys).
repeated(X,[_|Ys]) :-
	repeated(X,Ys).

listvars(X,V) :-
	listvars(X,V,[]).

listvars(X,[X|Y],Y) :-
	var(X),	
	!.
listvars(X,Y,Z) :-
	X =.. [_|Xs],
	arglistvars(Xs,Y,Z).

arglistvars([],Y,Y).
arglistvars([X|Xs],Y,Z) :-
	listvars(X,Y,W),
	arglistvars(Xs,W,Z).

newname(K,P) :-
	name(K,Kn),
	name(t,T),
	app(T,Kn,Tn),
	name(P,Tn).

check_fix([],[],[]).
check_fix([P|Ps],[change|Cs],[P|Chs]) :-
        check_fix(Ps,Cs,Chs).
check_fix([_|Ps],[nochange|Cs],Chs) :-
        check_fix(Ps,Cs,Chs).


check_fix([],[],[],[]).
check_fix([_|Ps],[RP|R],[RP1|S],Qs) :-
        fixpoint(RP,RP1),
	!,
        check_fix(Ps,R,S,Qs).
check_fix([P|Ps],[_|R],[_|S],[P|Qs]) :-
        check_fix(Ps,R,S,Qs).

fixpoint([],[]) :-
	!.
fixpoint(R1,R2) :-
	toptype(R1,T1),
	toptype(R2,T2),
	subtype(T2,T1,[R1|R2]).
	
toptype([proc(T1/1,_)|_],T1).

ub_proc([],_,R1,R1,nochange,K,K) :-  
        !.
ub_proc([P],R,[proc(TP/1,_)|_],D,nochange,K,K) :-  % ub with current def
	toptype(P,Q),
	subtype(Q,TP,R),	% no increase over previous def
	!,
	rdefs(TP,D,R).
ub_proc([P],R,[proc(TP/1,_)|_],D,change,K,K1) :-  % upper bound with current def
	!,
	toptype(P,Q),
	ub1(TP,Q,Q1,D1,R,K,K1),
	rdefs(Q1,D,[D1|R]).
ub_proc([P],R,_,D,change,K,K) :-  
	!,
	toptype(P,Q),
	rdefs(Q,D,R).
ub_proc([P1,P2|Ps],R,R1,U,Ch,K,K2) :-
	toptype(P1,T1),
	toptype(P2,T2),
	subtype(T1,T2,R),
	!,
	ub_proc([P2|Ps],[P2|R],R1,U,Ch,K,K2).
ub_proc([P1,P2|Ps],R,R1,U,Ch,K,K2) :-
	toptype(P1,T1),
	toptype(P2,T2),
	subtype(T2,T1,R),
	!,
	ub_proc([P1|Ps],[P1|R],R1,U,Ch,K,K2).
ub_proc([P1,P2|Ps],R,R1,U,Ch,K,K2) :-
	toptype(P1,T1),
	toptype(P2,T2),
        ub1(T1,T2,_,D,R,K,K1),
	ub_proc([D|Ps],[D|R],R1,U,Ch,K1,K2).

headclause(P/N,R,C,Xdef) :-
	functor(X,P,N),
	hclause(X,R,C,Xdef).

hclause(H,[[proc(T/1,[(TH :- B)])|Hdef]|_],(TH :- B),Hdef) :-
	TH =.. [T,H],
	!.
hclause(P,[_|Ps],Cl,Def) :-
	hclause(P,Ps,Cl,Def).


nested_memb(X,[X|_]).
nested_memb(X,[Y|_]) :-
	nested_memb(X,Y).
nested_memb(X,[_|Z]) :-
	nested_memb(X,Z).


typeclause(T/N,P,Cl) :-
	def(T/N,Q,P),
	!,
	memb2(Cl,Q).
	


merge_procs([],R,R).
merge_procs([P|Ps],R1,R2) :-
	insert_proc(P,R1,R3),
	merge_procs(Ps,R3,R2).

insert_proc(P,R,R) :-
	procmemb(P,R),
	!.
insert_proc(P,R,[P|R]).

procmemb(proc(F/N,_),[proc(F/N,_)|_]).
procmemb(P,[_|Ps]) :-
	procmemb(P,Ps).

:- dynamic traceon/0.

writestate(R) :-
	traceon,
	write(R),
	read(_),
	!,
	nl.
writestate(_) :-
	write('*'),
	sp_ttyflush.

lit(B, (B,_)).
lit(B, (_,Bs)) :-
        !,
        lit(B,Bs).
lit(B,B).

display_fix(Cls) :-
        nl,write('Enter filename: '),
        read(F),
        reg_filewrite(F,Cls).

reg_filewrite(tty,Cls) :-
        !,
	nl,write('Regular Approximation Program:'),nl,nl,
        displayqTypes(Cls),
	nl,nl.
reg_filewrite(F,Cls) :-
        tell(F),
        displayqTypes(Cls),
        told.

/* write the program to current output file  */

displayqTypes([[]|Ps]) :-
	displayqTypes(Ps).
displayqTypes([[Q|Qs]|Ps]) :-
	displayqPreddef(Q),
	displayqTypes(Qs),
	displayqTypes(Ps).
displayqTypes([proc(_,Qs)|Ps]) :-
	displayqRClauses(Qs),
	displayqTypes(Ps).
displayqTypes([]).

displayqPreddef(proc(_,[(H :- B)])) :-
	H =.. [_,A],
	displayqRClauses([(A :- B)]).


show_useless_clauses(P,F,K,K1) :-
	new_line,
	new_line,
	write('Find useless clauses (y/n)? '),
	getchar(YN),
	show_useless_clauses(YN,P,F,K,K1).

show_useless_clauses(y,P,F,K,K1) :-
	new_line,
	write('The following clauses are useless:'),
	new_line,
	new_line,
	!,
	useless_clause_check(P,F,K,K1,Us,[]),
	new_line,
	write('The following predicates are useless:'),
	new_line,
	write(Us).

show_useless_clauses(_,_,_,K,K).

useless_clause_check([],_,K,K,Us,Us).
useless_clause_check([P|Ps],R,K,K2,Us,Us2) :-
        user_clauses(P,Cls),
        check_bodies(P,1,Cls,R,K,K1,U),
	useless_pred(P,U,Us,Us1),
        useless_clause_check(Ps,R,K1,K2,Us1,Us2).

check_bodies(P,J,[(_ :- true)|Cls],R,K,K1,[J|U]) :-   
        !,
	J1 is J+1,
        check_bodies(P,J1,Cls,R,K,K1,U).
check_bodies(P,J,[(_ :- B)|Cls],R,K,K2,[J|U]) :-
        tsolve(B,R,T,Bdefs),
        unfold_body(T,Bdefs,T1),
        intersect_body(T1,Bdefs,_,_,K,K1),
        !,
	J1 is J+1,
        check_bodies(P,J1,Cls,R,K1,K2,U).
check_bodies(P,J,[(H :- B)|Cls],R,K,K1,U) :-          % if no solution in R.
	canonical((H :- B)),
%	write((H :- B)),
	write(P),
	write(': Clause '), 
	write(J),
	new_line,
	J1 is J+1,
        check_bodies(P,J1,Cls,R,K,K1,U).
check_bodies(_,_,[],_,K,K,[]).

useless_pred(P,[],[P|Us],Us) :-
	!.
useless_pred(_,_,Us,Us).
