% Convert a program to a unary predicate program
% This is the first stage in producing a regular unary program

% JPG: 25.6.93


% unary(F,S):  	F = source program file
%		S = list of argument types

unary(F,S) :-
	see(F),
	read(C),
	unary_clauses(C,S,[]),
	seen,
	!.


unary_clauses(X,S,S) :- 
	end_of_file(X),
	!.
unary_clauses((A :- B),S,S2) :- 
	!,
	canonical((A :- B)),
	arg_types(A,B,S,S1),
	read(C1),
	unary_clauses(C1,S1,S2).
unary_clauses(A,S,S2) :- 
	!,
	canonical(A),
	arg_types(A,true,S,S1),
	read(C1),
	unary_clauses(C1,S1,S2).

arg_types(A,B,S,S1) :-
	A =.. [P|Xs],
	functor(A,P,N),
	name(P,Pn),
	argtypenames(Pn,N,1,Ns),
	body_types(B,BT),
	arg_clauses(Xs,Ns,BT,S,S1).

argtypenames(_,M,N,[]) :-
	N > M,
	!.
argtypenames(Pn,K,M,[N|Ns]) :-
	name(M,Mn),
	my_append(Pn,[95,95|Mn],PMn),
	name(N,PMn),
	M1 is M+1,
	argtypenames(Pn,K,M1,Ns).

body_types(true,true) :-
	!.
body_types((B,Bs),G) :-
	sp_builtin(B),
	!,
	body_types(Bs,G).
body_types((B,Bs),G) :-
	!,
	B =.. [P|Xs],
        functor(B,P,N),
        name(P,Pn),
        argtypenames(Pn,N,1,Ns),
	body_arg_atoms(Xs,Ns,BG),
	body_types(Bs,G1),
	joingoals(BG,G1,G).
body_types(B,true) :-
	sp_builtin(B),
	!.
body_types(B,G) :-
	!,
	B =.. [P|Xs],
        functor(B,P,N),
        name(P,Pn),
        argtypenames(Pn,N,1,Ns),
	body_arg_atoms(Xs,Ns,G).

body_arg_atoms([],[],true).
body_arg_atoms([X],[N],A) :-
	!,
	A =.. [N,X].
body_arg_atoms([X|Xs],[N|Ns],(A,As)) :-
	A =.. [N,X],
	body_arg_atoms(Xs,Ns,As).

arg_clauses([],[],_,S,S).
arg_clauses([X|Xs],[N|Ns],BT,[(A :- BT)|S],S2) :-
	A =.. [N,X],
	arg_clauses(Xs,Ns,BT,S,S2).
	
