% Unfold uniform clause body 
% with respect to non-deterministic automaton program
% JPG - 20/8/98


unfold(Clause,R,UClause) :-
	unfold_step(Clause,R,Clause1),
	!,
	unfold(Clause1,R,UClause).
unfold(Clause,_,Clause).

unfold_step((H :- Body),R,(H :- Q)) :-
	nonvar_atom(Body, As,(A,Bs)),
	!,
	regular_clause(A,B,R),
	joingoals(B,Bs,Cs),
	joingoals(As,Cs,Q).
unfold_step((H :- Body),_,(H :- Body)) :-
	nonvar_arg(H),
	!.
unfold_step((H :- Body),R,(H :- Q)) :-
	headvar_atom(Body,As,(A,Bs)),
	!,
	regular_clause(A,B,R),
	joingoals(B,Bs,Cs),
	joingoals(As,Cs,Q).
unfold_step((H :- Body),_,(H :- Body)).

nonvar_atom((A,As),A,Bs) :-
	nonvar_arg(A),
	!,
	clause_tail(As,Bs).
nonvar_atom((A,Bs),(A,Cs),Ds) :-
	!,
	nonvar_atom(Bs,Cs,Ds).
nonvar_atom(A,true,(A,true)) :-
	nonvar_arg(A).

headvar_atom((A,As),A,Bs) :-
	nonvar_arg(A),
	!,
	clause_tail(As,Bs).
headvar_atom((A,Bs),(A,Cs),Ds) :-
	!,
	headvar_atom(Bs,Cs,Ds).
headvar_atom(A,true,(A,true)) :-
	nonvar_arg(A).

clause_tail((A,As),(A,As)) :-
	!.
clause_tail(A,(A,true)).

regular_clause(A,B,[(H :- Body)|_]) :-
	melt((H :- Body),(A :- B)).
regular_clause(A,B,[_|R]) :-
	regular_clause(A,B,R).

% reachability

reachable([S|Ss],R) :-
	t_reachable([S|Ss],R,_,_).

t_reachable([S|Ss],R,F,B) :-
	A =.. [S,X],
	regular_clause(A,true,R),	
	var(X),
	!,
	t_reachable(Ss,R,F,B).
t_reachable([S|Ss],R,F,B) :-
	A =.. [S,X],
	regular_clause(A,B,R),
	nonvar(X),
	X =.. [F|_],
	t_reachable(Ss,R,F,B).
t_reachable([],_,_,_).

% reachability (version 2)
% a set {s1,s2,...} is reachable if the goal s1(X),s2(X),... succeeds
% or if it has already been shown to be reachable.

reachable(Ss,Reached,Reached,_) :-
	memb2(Ss,Reached),
	!.
reachable(Ss,Reached,[Ss|Reached],R) :-
	reach_goal(Ss,G),
	succeeds(G,R),
	!.

succeeds(true,_) :-
	!.
succeeds((G,Gs),R) :-
	!,
	succeeds(G,R),
	succeeds(Gs,R).
succeeds(G,R) :-
	regular_clause(G,B,R),
	succeeds(B,R).
	
reachability_sets(true,[]) :-
	!.
reachability_sets((A,As),RSets) :-
	!,
	A =.. [P,X],
	reachability_sets(As,RSets1),
	insert_reachable_predset(P,X,RSets1,RSets).
reachability_sets(P,X,[rset(X,[P])]).

insert_reachable_predset(P,X,[rset(Y,Ps)|Rset],[rset(Y,Ps1)|Rset]) :-
	X == Y,
	!,
	insert_reachable_pred(P,Ps,Ps1).
insert_reachable_predset(P,X,[rset(Y,Ps)|Rset],[rset(Y,Ps)|Rset1]) :-
	insert_reachable_predset(P,X,Rset,Rset1).
insert_reachable_predset(P,X,[],[rset(X,[P])]).
	
%insert_reachable_pred(P,Ps,Ps) :-
%	memb2(P,Ps),
%	!.
%insert_reachable_pred(P,Ps,[P|Ps]).

insert_reachable_pred(P,[],[P]) :-
	!.
insert_reachable_pred(P,[P|Ps],[P|Ps]) :-
	!.
insert_reachable_pred(P,[Q|Qs],[Q|Ps]) :-
	P @> Q,
	!,
	insert_reachable_pred(P,Qs,Ps).
insert_reachable_pred(P,[Q|Ps],[P,Q|Ps]).

testprog([
	(p(f('X0')) :- q('X0')),
	(s(f('X0')) :- r('X0')),
	(t(f('X0')) :- p('X0'),s('X0')),
	(q(a) :- true),
	(r(b) :- true),
	(r(a) :- true)]).

reach_goal(Ss,G) :-
	reach_goal(Ss,G,_).

reach_goal([P],A,X) :-
	A =.. [P,X].
reach_goal([P,Q|Ps],(A,As),X) :-
	A =.. [P,X],
	reach_goal([Q|Ps],As,X).
